r"""ABC Index descriptor.

References
    * http://match.pmf.kg.ac.rs/electronic_versions/Match75/n1/match75n1_233-242.pdf

"""

import numpy as np

from ._base import Descriptor
from ._graph_matrix import DistanceMatrix

__all__ = ("ABCIndex", "ABCGGIndex")


class ABCIndexBase(Descriptor):
    __slots__ = ()

    @classmethod
    def preset(cls, version):
        yield cls()

    explicit_hydrogens = False

    def parameters(self):
        return ()

    def __str__(self):
        return self.__class__.__name__[:-5]

    rtype = float


class ABCIndex(ABCIndexBase):
    r"""atom-bond connectivity index descriptor.

    References:
        * :doi:`10.2298/FIL1204733D`

    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "atom-bond connectivity index"

    @staticmethod
    def _each_bond(bond):
        du = bond.GetBeginAtom().GetDegree()
        dv = bond.GetEndAtom().GetDegree()

        return np.sqrt((du + dv - 2.0) / (du * dv))

    def calculate(self):
        return float(sum(self._each_bond(bond) for bond in self.mol.GetBonds()))


class ABCGGIndex(ABCIndexBase):
    r"""Graovac-Ghorbani atom-bond connectivity index descriptor.

    References:
        * Furtula, B. Atom-bond connectivity index versus Graovac-Ghorbani analog. MATCH Commun. Math. Comput. Chem 75, 233-242 (2016).

    """  # noqa: E501

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "Graovac-Ghorbani atom-bond connectivity index"

    def dependencies(self):
        return {"D": DistanceMatrix(self.explicit_hydrogens)}

    @staticmethod
    def _each_bond(bond, D):
        u = bond.GetBeginAtomIdx()
        v = bond.GetEndAtomIdx()

        nu = np.sum(D[u, :] < D[v, :])
        nv = np.sum(D[v, :] < D[u, :])

        return np.sqrt((nu + nv - 2.0) / (nu * nv))

    def calculate(self, D):
        return float(sum(self._each_bond(bond, D) for bond in self.mol.GetBonds()))



r"""Wildman-Crippen LogP/MR descriptor.

References
    * :doi:`10.1021/ci990307l`

"""
from rdkit.Chem import Crippen

from ._base import Descriptor

__all__ = ("SLogP", "SMR")


class WildmanCrippenBase(Descriptor):
    __slots__ = ()

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return self.__class__.__name__

    def parameters(self):
        return ()

    explicit_hydrogens = False

    rtype = float


class SLogP(WildmanCrippenBase):
    r"""Wildman-Crippen LogP descriptor(rdkit wrapper)."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "Wildman-Crippen LogP"

    def calculate(self):
        return Crippen.MolLogP(self.mol)


class SMR(WildmanCrippenBase):
    r"""Wildman-Crippen MR descriptor(rdkit wrapper)."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "Wildman-Crippen MR"

    def calculate(self):
        return Crippen.MolMR(self.mol)



import os
from glob import glob

import yaml
import numpy as np
from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred import Calculator, Polarizability, descriptors
from mordred.error import MissingValueBase

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


data_dir = os.path.join(os.path.dirname(__file__), "references")


def test_by_references():
    calc = Calculator(
        d
        for d in descriptors.all
        if d.__class__ not in [Polarizability.APol, Polarizability.BPol]
    )

    calc.register([Polarizability.APol(True), Polarizability.BPol(True)])

    actuals = {}
    for mol in Chem.SDMolSupplier(
        os.path.join(data_dir, "structures.sdf"), removeHs=False
    ):
        actuals[mol.GetProp("_Name")] = {
            str(d): v for d, v in zip(calc.descriptors, calc(mol))
        }

    for path in glob(os.path.join(data_dir, "*.yaml")) + glob(
        os.path.join(data_dir, "**/*.yaml")
    ):
        for test in yaml.load(open(path), Loader=Loader):
            dnames = test["names"]
            if not isinstance(dnames, list):
                dnames = [dnames]

            desireds = (
                (mname, zip(dnames, values if isinstance(values, list) else [values]))
                for mname, values in test["results"].items()
            )

            digit = test.get("digit")
            if digit is None:
                digit = 0  # exact match

            def assert_f(a, d, m):
                if np.isnan(d):
                    assert isinstance(a, MissingValueBase)
                    return

                assert_almost_equal(a, d, digit, m)

            for mname, descs in desireds:
                for dname, desired in descs:
                    if not desired == "skip":
                        assert_f(
                            actuals[mname][dname],
                            desired,
                            "{} of {}".format(dname, mname),
                        )



from __future__ import print_function

import os
import sys

import numpy as np


def parse_enum(enum, v):
    if isinstance(v, enum):
        return v
    else:
        return enum[v]


def atoms_to_numpy(f, mol, dtype="float"):
    return np.fromiter((f(a) for a in mol.GetAtoms()), dtype, mol.GetNumAtoms())


def conformer_to_numpy(conf):
    return np.array([list(conf.GetAtomPosition(i)) for i in range(conf.GetNumAtoms())])


class Capture(object):
    def __init__(self, target="stderr"):
        self.target = target
        self.orig = getattr(sys, target)
        self.result = []

    def write(self, text):
        self.result.append(text)

    def flush(self):
        pass

    def __enter__(self):
        setattr(sys, self.target, self)
        return self

    def __exit__(self, *args):
        setattr(sys, self.target, self.orig)


class DummyBar(object):
    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args, **kws):
        pass

    def update(self, *args, **kws):
        pass

    @classmethod
    def write(cls, s, file=sys.stdout, end="\n"):
        print(s, file=file, end=end)  # noqa: T003


class NotebookWrapper(object):
    def __init__(self, **kwargs):
        from tqdm import tqdm_notebook

        self.bar = tqdm_notebook(**kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def update(self, *args, **kwargs):
        self.bar.update(*args, **kwargs)

    def write(self, *args, **kwargs):
        self.bar.update(*args, **kwargs)


def PathType(string):
    if not os.path.isfile(string):
        raise ValueError("file not exists: {}".format(string))

    return string


def module_prog(pkg):
    return "{} -m {}".format(os.path.basename(sys.executable), pkg)


def to_ordinal(n):
    r"""Int to ordinal string.

    >>> to_ordinal(1)
    'first'
    >>> to_ordinal(2)
    'second'
    >>> to_ordinal(3)
    'third'
    >>> to_ordinal(4)
    '4-th'
    >>> to_ordinal(104)
    '104-th'

    """
    if n == 1:
        return "first"
    elif n == 2:
        return "second"
    elif n == 3:
        return "third"
    else:
        return "{}-th".format(n)



import inspect
import operator
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from packaging.version import Version as StrictVersion

import six
import numpy as np
from rdkit import Chem

if hasattr(inspect, "getfullargspec"):

    def getargs(func):
        return tuple(inspect.getfullargspec(func).args[1:])

else:

    def getargs(func):
        try:
            return tuple(inspect.getargspec(func).args[1:])
        except TypeError:
            return ()


class MissingValueException(Exception):
    """Internally used exception."""

    __slots__ = ("error",)

    def __init__(self, error):
        self.error = error


class DescriptorMeta(ABCMeta):
    def __new__(cls, classname, bases, dict):
        __init__ = dict.get("__init__")
        if __init__ is None:
            for base in bases:
                __init__ = getattr(base, "__init__", None)
                if __init__ is not None:
                    break

        dict["parameter_names"] = getargs(__init__)
        if "since" in dict:
            dict["since"] = StrictVersion(dict["since"])

        return ABCMeta.__new__(cls, classname, bases, dict)


class Descriptor(six.with_metaclass(DescriptorMeta, object)):
    r"""Abstract base class of descriptors.

    Attributes:
        mol(rdkit.Mol): target molecule

    """

    __slots__ = ("_context",)

    explicit_hydrogens = True
    kekulize = False
    require_connected = False
    require_3D = False

    def __reduce_ex__(self, version):
        return self.__class__, self.parameters()

    def description(self):
        pass

    @classmethod
    def preset(cls, version):
        r"""Generate preset descriptor instances.

        Returns:
            Iterable[Descriptor]: preset descriptors

        """
        return ()

    @abstractmethod
    def parameters(self):
        """[abstractmethod] get __init__ arguments of this descriptor instance.

        this method used in pickling and identifying descriptor instance.

        Returns:
            tuple: tuple of __init__ arguments

        """
        raise NotImplementedError("not implemented Descriptor.parameters method")

    def get_parameter_dict(self):
        return dict(zip(self.parameter_names, self.parameters()))

    def to_json(self):
        """Convert to json serializable dictionary.

        Returns:
            dict: dictionary of descriptor

        """
        d, ps = self._to_json()
        if len(ps) == 0:
            return {"name": d}
        else:
            return {"name": d, "args": ps}

    def _to_json(self):
        d = self.__class__.__name__
        ps = self.get_parameter_dict()

        return d, {k: getattr(v, "as_argument", v) for k, v in ps.items()}

    @abstractmethod
    def calculate(self):
        r"""[abstractmethod] calculate descriptor value.

        Returns:
            rtype

        """
        raise NotImplementedError("not implemented Descriptor.calculate method")

    def dependencies(self):
        r"""Descriptor dependencies.

        Returns:
            dict[str, Descriptor or None] or None

        """
        pass

    @property
    def as_argument(self):
        """Argument representation of descriptor.

        Returns:
            any

        """
        return self

    @staticmethod
    def _pretty(v):
        v = getattr(v, "as_argument", v)
        return repr(v)

    def __repr__(self):
        return "{}.{}({})".format(
            self.__class__.__module__,
            self.__class__.__name__,
            ", ".join(self._pretty(a) for a in self.parameters()),
        )

    def __hash__(self):
        return hash((self.__class__, self.parameters()))

    def __compare_by_reduce(meth):
        def compare(self, other):
            L = self.__class__, self.parameters()
            r = other.__class__, other.parameters()
            return getattr(L, meth)(r)

        return compare

    __eq__ = __compare_by_reduce("__eq__")
    __ne__ = __compare_by_reduce("__ne__")

    __lt__ = __compare_by_reduce("__lt__")
    __gt__ = __compare_by_reduce("__gt__")
    __le__ = __compare_by_reduce("__le__")
    __ge__ = __compare_by_reduce("__ge__")

    rtype = None

    @property
    def mol(self):
        """Get molecule.

        Returns:
            rdkit.Mol

        """
        return self._context.get_mol(self)

    @property
    def config(self):
        return self._context.config

    @property
    def coord(self):
        """Get 3D coordinate.

        Returns:
            numpy.array[3, N]: coordinate matrix

        """
        if not self.require_3D:
            self.fail(AttributeError("use 3D coordinate in 2D descriptor"))

        return self._context.get_coord(self)

    def get_3D_mol(self):
        mol = Chem.Mol(self.mol)
        conf = Chem.Conformer(mol.GetNumAtoms())
        for i, xyz in enumerate(self.coord):
            conf.SetAtomPosition(i, xyz)

        mol.AddConformer(conf)
        return mol

    def fail(self, exception):
        """Raise known exception and return missing value.

        Raises:
            MissingValueException

        """
        raise MissingValueException(exception)

    @contextmanager
    def rethrow_zerodiv(self):
        """[contextmanager] treat zero div as known exception."""
        with np.errstate(divide="raise", invalid="raise"):
            try:
                yield
            except (FloatingPointError, ZeroDivisionError) as e:
                self.fail(ZeroDivisionError(*e.args))

    @contextmanager
    def rethrow_na(self, exception):
        """[contextmanager] treat any exceptions as known exception."""
        try:
            yield
        except exception as e:
            self.fail(e)

    def _unary_common(name, operator):
        def unary(self):
            return UnaryOperatingDescriptor(name.format(self), operator, self)

        return unary

    def _binary_common(name, operator):
        def binary(self, other):
            if not isinstance(other, Descriptor):
                other = ConstDescriptor(other)

            return BinaryOperatingDescriptor(
                name.format(self, other), operator, self, other
            )

        return binary

    __add__ = _binary_common("({}+{})", "+")
    __sub__ = _binary_common("({}-{})", "-")
    __mul__ = _binary_common("({}*{})", "*")
    __truediv__ = _binary_common("({}/{})", "/")
    __floordiv__ = _binary_common("({}//{})", "//")
    __mod__ = _binary_common("({}%{})", "%")
    __pow__ = _binary_common("({}**{})", "**")

    __neg__ = _unary_common("-{}", "-")
    __pos__ = _unary_common("+{}", "+")
    __abs__ = _unary_common("|{}|", "abs")
    __trunc__ = _unary_common("trunc({})", "trunc")

    if six.PY3:
        __ceil__ = _unary_common("ceil({})", "ceil")
        __floor__ = _unary_common("floor({})", "floor")


def is_descriptor_class(desc, include_abstract=False):
    r"""Check calculatable descriptor class or not.

    Returns:
        bool

    """
    return (
        isinstance(desc, type)
        and issubclass(desc, Descriptor)
        and (True if include_abstract else not inspect.isabstract(desc))
    )


class UnaryOperatingDescriptor(Descriptor):
    @classmethod
    def preset(cls, version):
        return cls()

    operators = {
        "+": operator.pos,
        "-": operator.neg,
        "abs": operator.abs,
        "trunc": np.trunc,
        "ceil": np.ceil,  # noqa: S001
        "floor": np.floor,
    }

    def parameters(self):
        return self._name, self._operator, self._value

    def __init__(self, name, operator, value):
        self._name = name
        self._operator = operator
        self._fn = self.operators[operator]
        self._value = value

    def _to_json(self):
        return (
            self.__class__.__name__,
            {
                "name": self._name,
                "operator": self._operator,
                "value": self._value.to_json(),
            },
        )

    def __str__(self):
        return self._name

    def dependencies(self):
        return {"value": self._value}

    def calculate(self, value):
        return self._fn(value)


class ConstDescriptor(Descriptor):
    @classmethod
    def preset(cls, version):
        return cls()

    def parameters(self):
        return (self._value,)

    def __init__(self, value):
        self._value = value

    def __str__(self):
        return str(self._value)

    def calculate(self):
        return self._value


class BinaryOperatingDescriptor(Descriptor):
    @classmethod
    def preset(cls, version):
        return cls()

    operators = {
        "+": operator.add,
        "-": operator.sub,
        "*": operator.mul,  # noqa: S001
        "/": operator.truediv,
        "//": operator.floordiv,
        "%": operator.mod,  # noqa: S001
        "**": operator.pow,
    }

    def _to_json(self):
        return (
            self.__class__.__name__,
            {
                "name": self._name,
                "operator": self._operator,
                "left": self._left.to_json(),  # noqa: S001
                "right": self._right.to_json(),
            },
        )

    def parameters(self):
        return self._name, self._operator, self._left, self._right

    def __init__(self, name, operator, left, right):
        self._name = name
        self._operator = operator
        self._fn = self.operators[operator]
        self._left = left
        self._right = right

    def __str__(self):
        return self._name

    def dependencies(self):
        return {"left": self._left, "right": self._right}

    def calculate(self, left, right):
        return self._fn(left, right)



from ..error import MissingValueBase


def is_missing(v):
    """Check argument is either MissingValue or not.

    >>> from mordred.error import Missing, Error
    >>> is_missing(1)
    False
    >>> is_missing(Missing(ValueError("missing"), ()))
    True
    >>> is_missing(Error(ValueError("error"), ()))
    True

    Parameters:
        v(any): value

    Returns:
        bool

    """
    return isinstance(v, MissingValueBase)



from __future__ import division

from collections import defaultdict

from rdkit.Chem import HybridizationType
from rdkit.Chem.rdMolDescriptors import CalcFractionCSP3

from ._base import Descriptor

__all__ = ("CarbonTypes", "HybridizationRatio", "FractionCSP3")


class CarbonTypesBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False
    kekulize = True


class CarbonTypesCache(CarbonTypesBase):
    __slots__ = ()

    def parameters(self):
        return ()

    _hybridization = {
        HybridizationType.SP: 1,
        HybridizationType.SP2: 2,
        HybridizationType.SP3: 3,
        HybridizationType.SP3D: 3,
        HybridizationType.SP3D2: 3,
    }

    def calculate(self):
        r = defaultdict(lambda: defaultdict(int))
        for a in self.mol.GetAtoms():
            if a.GetAtomicNum() != 6:
                continue

            carbon = sum(other.GetAtomicNum() == 6 for other in a.GetNeighbors())

            SP = self._hybridization.get(a.GetHybridization())

            r[SP][carbon] += 1

        return r


class CarbonTypes(CarbonTypesBase):
    r"""carbon types descriptor.

    :type nCarbon: int
    :param nCarbon: count `n`-carbon bonded carbon

    :type SP: int
    :param SP: count :math:`{\rm SP}n` carbon
    """

    since = "1.0.0"
    __slots__ = ("_nCarbon", "_SP")

    def description(self):
        return "SP{} carbon bound to {} other carbon{}".format(
            self._SP if self._SP != 1 else "",
            self._nCarbon,
            "s" if self._nCarbon > 1 else "",
        )

    @classmethod
    def preset(cls, version):
        return map(
            lambda args: cls(*args),
            [(1, 1), (2, 1), (1, 2), (2, 2), (3, 2), (1, 3), (2, 3), (3, 3), (4, 3)],
        )

    def __str__(self):
        return "C{}SP{}".format(self._nCarbon, self._SP)

    def parameters(self):
        return self._nCarbon, self._SP

    def __init__(self, nCarbon=1, SP=3):
        assert SP in [1, 2, 3]

        self._nCarbon = nCarbon
        self._SP = SP

    def dependencies(self):
        return {"CT": CarbonTypesCache()}

    def calculate(self, CT):
        return CT[self._SP][self._nCarbon]

    rtype = int


class HybridizationRatio(CarbonTypesBase):
    r"""hybridization ratio descriptor.

    .. math::

        {\rm HybRatio} = \frac{N_{\rm SP3}}{N_{\rm SP2} + N_{\rm SP3}}

    :returns: NaN when :math:`N_{\rm SP2} + N_{\rm SP3} = 0`.
    """

    def description(self):
        return "hybridization ratio"

    since = "1.0.0"
    __slots__ = ()

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return "HybRatio"

    def parameters(self):
        return ()

    def dependencies(self):
        return {"CT": CarbonTypesCache()}

    def calculate(self, CT):
        Nsp3 = sum(CT[3].values())
        Nsp2 = sum(CT[2].values())

        if Nsp3 == Nsp2 == 0:
            self.fail(ValueError("there are no sp3 and sp2 carbons"))

        return Nsp3 / (Nsp2 + Nsp3)

    rtype = float


class FractionCSP3(Descriptor):
    r"""the fraction of C atoms that are SP3 hybridized."""

    __slots__ = ()
    since = "1.1.0"

    @classmethod
    def preset(cls, version):
        yield cls()

    def description(self):
        return "the fraction of C atoms that are SP3 hybridized"

    def __str__(self):
        return "FCSP3"

    def parameters(self):
        return ()

    def calculate(self):
        return CalcFractionCSP3(self.mol)

    rtype = float



import numpy as np
from six import string_types, integer_types

from .util import is_missing
from .descriptor import Descriptor


class Result(object):
    r"""Result type."""

    __slots__ = ("mol", "_values", "_descriptors", "_name_to_value")

    def __init__(self, mol, r, d):
        self.mol = mol
        self._values = list(r)
        self._descriptors = list(d)
        self._name_to_value = None

    def __str__(self):
        return "{}({{{}}})".format(
            self.__class__.__name__,
            ", ".join(
                "'{}': {}".format(k, v) for k, v in zip(self._descriptors, self._values)
            ),
        )

    def __repr__(self):
        return "{}({!r},{!r},{!r})".format(
            self.__class__.__name__, self.mol, self._values, self._descriptors
        )

    def fill_missing(self, value=np.nan):
        r"""Replace missing value to "value".

        Parameters:
            value: value that missing value is replaced

        Returns:
            Result

        """
        return self.__class__(
            self.mol,
            [(value if is_missing(v) else v) for v in self.values()],
            self.keys(),
        )

    def drop_missing(self):
        r"""Delete missing value.

        Returns:
            Result

        """
        newvalues = []
        newdescs = []
        for d, v in self.items():
            if not is_missing(v):
                newvalues.append(v)
                newdescs.append(d)

        return self.__class__(self.mol, newvalues, newdescs)

    def items(self):
        r"""Get items.

        Returns:
            Iterable[(Descriptor, value)]

        """
        return ((k, v) for k, v in zip(self.keys(), self.values()))

    def keys(self):
        r"""Get descriptors instances.

        Returns:
            Iterable[Descriptor]

        """
        return iter(self._descriptors)

    def values(self):
        r"""Get descriptor values.

        Returns:
            Iterable[value]

        """
        return iter(self._values)

    __iter__ = values

    def __reversed__(self):
        return reversed(self._values)

    def asdict(self, rawkey=False):
        r"""Convert Result to dict.

        Parameters:
            rawkey(bool):
                * True: dict key is Descriptor instance
                * False: dict key is str

        Returns:
            dict

        """
        if rawkey:
            return dict(self.items())
        else:
            return {str(k): v for k, v in self.items()}

    @property
    def ix(self):
        r"""Access descriptor value by index.

        >>> from mordred import Calculator, Lipinski
        >>> from rdkit import Chem
        >>> result = Calculator(Lipinski.Lipinski)(Chem.MolFromSmiles("C1CCCCC1"))
        >>> result.ix[0]
        True
        """
        return GetValueByIndex(self._values)

    @property
    def name(self):
        r"""Access descriptor value by descriptor name or instance.

        >>> from mordred import Calculator, descriptors
        >>> from rdkit import Chem
        >>> result = Calculator(descriptors)(Chem.MolFromSmiles("C1CCCCC1"))
        >>> result.name["C2SP3"]
        6

        """
        if self._name_to_value is None:
            self._name_to_value = {
                str(d): v for d, v in zip(self._descriptors, self._values)
            }

        return GetValueByName(self._name_to_value)

    def __getitem__(self, key):
        if isinstance(key, (integer_types, slice)):
            return self.ix[key]

        elif isinstance(key, (string_types, Descriptor)):
            return self.name[key]

        else:
            raise TypeError(
                "Result indices must be "
                "integers, slices, strings or Descriptor instance, "
                "not {}".format(key.__class__.__name__)
            )

    def __len__(self):
        return len(self._descriptors)


class GetValueByIndex(object):
    __slots__ = ("_values",)

    def __init__(self, values):
        self._values = values

    def __getitem__(self, key):
        return self._values[key]


class GetValueByName(object):
    __slots__ = ("_name_to_value",)

    def __init__(self, name_to_value):
        self._name_to_value = name_to_value

    def __getitem__(self, key):
        return self._name_to_value[str(key)]



import math
import operator

import six
from rdkit import Chem


from .. import ABCIndex

binary = [
    operator.add,
    operator.sub,
    operator.mul,
    operator.truediv,
    operator.floordiv,
    operator.mod,
    operator.pow,
]

unary = [operator.neg, operator.pos, operator.abs, math.trunc]


if six.PY3:
    unary.extend([math.ceil, math.floor])


def test_compose_descriptor():
    L = ABCIndex.ABCIndex()
    r = ABCIndex.ABCGGIndex()
    m = Chem.MolFromSmiles("c1ccccc1C")

    for op in binary:
        assert op(L, r)(m) == op(L(m), r(m))

    for op in unary:
        assert op(L)(m) == op(L(m))



from __future__ import division

from ._base import Descriptor
from ._graph_matrix import Radius3D as CRadius3D
from ._graph_matrix import Diameter3D as CDiameter3D

__all__ = ("Diameter3D", "Radius3D", "GeometricalShapeIndex", "PetitjeanIndex3D")


class GeometricalIndexBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = True
    require_3D = True

    @classmethod
    def preset(cls, version):
        yield cls()

    def parameters(self):
        return ()

    rtype = float


class Radius3D(GeometricalIndexBase):
    r"""geometric radius descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "geometric radius"

    def __str__(self):
        return "GeomRadius"

    def dependencies(self):
        return {"R": CRadius3D(self.explicit_hydrogens)}

    def calculate(self, R):
        return R


class Diameter3D(GeometricalIndexBase):
    r"""geometric diameter descriptor."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "geometric diameter"

    def __str__(self):
        return "GeomDiameter"

    def dependencies(self):
        return {"D": CDiameter3D(self.explicit_hydrogens)}

    def calculate(self, D):
        return D


class GeometricalShapeIndex(GeometricalIndexBase):
    r"""geometrical shape index descriptor.

    .. math::

        I_{\rm topo} = \frac{D - R}{R}

    where
    :math:`R` is geometric radius,
    :math:`D` is geometric diameter.

    :returns: NaN when :math:`R = 0`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "geometrical shape index"

    def __str__(self):
        return "GeomShapeIndex"

    def dependencies(self):
        return {
            "D": CDiameter3D(self.explicit_hydrogens),
            "R": CRadius3D(self.explicit_hydrogens),
        }

    def calculate(self, R, D):
        with self.rethrow_zerodiv():
            return (D - R) / (R)


class PetitjeanIndex3D(GeometricalShapeIndex):
    r"""geometric Petitjean index descriptor.

    .. math::

        I_{\rm Petitjean} = \frac{D - R}{D}

    where
    :math:`R` is geometric radius,
    :math:`D` is geometric diameter.

    :returns: NaN when :math:`D = 0`
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "geometric Petitjean index"

    def __str__(self):
        return "GeomPetitjeanIndex"

    def calculate(self, R, D):
        with self.rethrow_zerodiv():
            return (D - R) / (D)



import os
import sys
import shutil
import tempfile
from contextlib import contextmanager

from rdkit.Chem import AllChem as Chem


from .. import Calculator, descriptors
from ..__main__ import main as mordred

Nd2D = len(Calculator(descriptors, ignore_3D=True).descriptors)
Nd3D = len(Calculator(descriptors, ignore_3D=False).descriptors)

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

__version__ = version("mordredcommunity")


def in_(a, s):
    assert a in s, "{!r} not in {!r}".format(a, s)


@contextmanager
def isolate():
    cwd = os.getcwd()
    try:
        work = tempfile.mkdtemp()
        os.chdir(work)
        yield
    finally:
        os.chdir(cwd)
        shutil.rmtree(work)


class Result(object):
    def __iter__(self):
        return iter([self.stdout, self.stderr, self.exitcode])

    def __repr__(self):
        return "{!r} {!r} {!r}".format(self.stdout, self.stderr, self.exitcode)


def command(cmd, *args):
    r = Result()

    orig = sys.stdout, sys.stderr

    ofd, opath = tempfile.mkstemp()
    efd, epath = tempfile.mkstemp()

    try:
        with os.fdopen(ofd, "w") as stdout, os.fdopen(efd, "w") as stderr:
            sys.stdout = stdout
            sys.stderr = stderr

            try:
                cmd(args=args)
                r.exitcode = 0
            except SystemExit as e:
                r.exitcode = e.args[0]

    finally:
        sys.stdout, sys.stderr = orig

        with open(opath) as stdout, open(epath) as stderr:
            r.stdout = stdout.read()
            r.stderr = stderr.read()

        os.remove(opath)
        os.remove(epath)

    return r


def test_no_args():
    stdout, stderr, exitcode = command(mordred)
    assert exitcode == 2
    assert stdout == ""
    in_("usage:", stderr)
    # python3 or python2
    assert (
        "the following arguments are required: INPUT" in stderr
        or "too few arguments" in stderr
    )


def test_help():
    stdout, stderr, exitcode = command(mordred, "-h")
    assert exitcode == 0
    assert stderr == ""
    in_("usage:", stdout)
    in_("descriptors:", stdout)


def test_version():
    stdout, stderr, exitcode = command(mordred, "--version")
    assert exitcode == 0

    vstr = "mordred-{}\n".format(__version__)

    assert stdout == vstr


def test_missing_file():
    with isolate():
        stdout, stderr, exitcode = command(mordred, "missing.smi")
        assert exitcode == 2
        assert stdout == ""
        in_("usage:", stderr)
        in_("invalid PathType value", stderr)


def number_of_field(output):
    return len(list(filter(lambda c: c == ",", output.split("\n")[0])))


def test_smi():
    with isolate():
        with open("input.smi", "w") as f:
            f.write("c1ccccc1 Benzene\n")

        stdout, stderr, exitcode = command(mordred, "input.smi", "-q", "-o", "-")

        assert exitcode == 0
        assert number_of_field(stdout) == Nd2D
        assert stdout.split("\n")[1].split(",")[0] == "Benzene"
        assert stderr == ""


def test_smi_without_name():
    with isolate():
        with open("input.smi", "w") as f:
            f.write("c1ccccc1\n")

        stdout, stderr, exitcode = command(mordred, "input.smi", "-q", "-o", "-")

        assert exitcode == 0
        assert number_of_field(stdout) == Nd2D
        assert stdout.split("\n")[1].split(",")[0] == "c1ccccc1"
        assert stderr == ""


def test_sdf():
    with isolate():
        mol = Chem.MolFromSmiles("c1ccccc1")
        mol.SetProp("_Name", "Benzene")
        Chem.MolToMolFile(mol, "input.sdf")

        stdout, stderr, exitcode = command(
            mordred, "input.sdf", "-q", "-o", "output.csv"
        )

        assert exitcode == 0
        assert stdout == ""
        assert stderr == ""
        output = open("output.csv").read()
        assert number_of_field(output) == Nd2D
        assert output.split("\n")[1].split(",")[0] == "Benzene"


def test_sdf_3D():
    with isolate():
        mol = Chem.MolFromSmiles("c1ccccc1")
        Chem.EmbedMolecule(mol)
        mol.SetProp("_Name", "Benzene")
        Chem.MolToMolFile(mol, "input.sdf")

        stdout, stderr, exitcode = command(
            mordred, "input.sdf", "-q", "-o", "output.csv", "-3"
        )

        assert exitcode == 0
        assert stdout == ""
        assert stderr == ""
        output = open("output.csv").read()
        assert number_of_field(output) == Nd3D
        assert output.split("\n")[1].split(",")[0] == "Benzene"


def test_verbose():
    with isolate():
        with open("input.smi", "w") as f:
            f.write("c1ccccc1 Benzene\n")

        r = command(mordred, "input.smi", "-o", "-", "-q", "-vv")

        assert r.exitcode == 0
        assert number_of_field(r.stdout) == Nd2D
        assert r.stdout.split("\n")[1].split(",")[0] == "Benzene"
        assert "[Missing]" in r.stderr



import os

from rdkit import Chem

from mordred.CPSA import TotalSurfaceArea
from mordred.surface_area import SurfaceArea

# calculated by pymol
txt_data = """
Hexane               296.910
Benzene              243.552
Caffeine             369.973
Cyanidin             483.873
Lycopene            1172.253
Epicatechin          489.498
Limonene             361.278
Allicin              356.872
Glutathione          530.679
Digoxin             1074.428
Capsaicin            641.527
EllagicAcid          440.267
Astaxanthin         1080.941
DMSO                 227.926
DiethylThioketone    290.503
VinylsulfonicAcid    246.033
Thiophene            227.046
Triethoxyphosphine   396.482
MethylphosphonicAcid 235.685
MethylCyclopropane   229.071
Acetonitrile         182.197
Histidine            335.672
"""[
    1:-1
]


sdf_file = os.path.join(os.path.dirname(__file__), "references", "structures.sdf")


def test_SASA():
    data = {}
    for line in txt_data.split("\n"):
        n, v = line.strip().split()
        data[n] = float(v)

    tsa = TotalSurfaceArea()

    for mol in Chem.SDMolSupplier(sdf_file, removeHs=False):
        name = mol.GetProp("_Name")
        actual = sum(SurfaceArea.from_mol(mol).surface_area())

        if name not in data:
            continue

        desired = data[name]

        e = abs((actual - desired) / desired)
        p = 0.05

        assert e < p, "large SASA error in {}: {}".format(name, e)

        e = abs((tsa(mol) - desired) / desired)

        assert e < p, "large SASA error in {}: {}".format(name, e)



import networkx
from rdkit import Chem

from ._base import Descriptor

__all__ = ("RingCount",)


class RingCountBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False

    def parameters(self):
        return ()


class Rings(RingCountBase):
    __slots__ = ()

    def calculate(self):
        return [frozenset(s) for s in Chem.GetSymmSSSR(self.mol)]


class FusedRings(RingCountBase):
    __slots__ = ()

    def dependencies(self):
        return {"Rings": Rings()}

    def calculate(self, Rings):
        if len(Rings) < 2:
            return []

        G = networkx.Graph()

        L = len(Rings)
        for i, j in ((i, j) for i in range(L) for j in range(i + 1, L)):
            if len(Rings[i] & Rings[j]) >= 2:
                G.add_edge(i, j)

        return [
            frozenset(j for i in ring_ids for j in Rings[i])
            for ring_ids in networkx.connected_components(G)
        ]


class RingCount(RingCountBase):
    r"""ring count descriptor.

    :type order: int or None
    :param order: number of bonds in ring

    :type greater: bool
    :param greater: count length or greater rings

    :type fused: bool
    :param fused: count fused rings

    :type aromatic: bool or None
    :param aromatic:
        * True: count aromatic rings
        * False: count non-aromatic rings
        * None: count any rings

    :type hetero: bool or None
    :param hetero:
        * True: count hetero rings
        * False: count carbon rings
        * None: count any rings
    """

    since = "1.0.0"
    __slots__ = ("_order", "_greater", "_fused", "_aromatic", "_hetero")

    def description(self):
        if self._order is None:
            o = ""
        elif self._greater:
            o = "{}-or-greater-membered ".format(self._order)
        else:
            o = "{}-membered ".format(self._order)

        if self._aromatic is None:
            a = ""
        elif self._aromatic is True:
            a = "aromatic "
        else:
            a = "aliphatic "

        return "{}{}{}{}ring count".format(
            o, a, "fused " if self._fused else "", "hetero " if self._hetero else ""
        )

    @classmethod
    def preset(cls, version):
        for fused in [False, True]:
            for arom in [None, True, False]:
                for hetero in [None, True]:
                    yield cls(None, False, fused, arom, hetero)
                    for n in range(4 if fused else 3, 13):
                        yield cls(n, False, fused, arom, hetero)

                    yield cls(12, True, fused, arom, hetero)

    def __str__(self):
        attrs = []

        if self._greater:
            attrs.append("G")

        if self._order is not None:
            attrs.append(str(self._order))

        if self._fused:
            attrs.append("F")

        if self._aromatic is True:
            attrs.append("a")
        elif self._aromatic is False:
            attrs.append("A")

        if self._hetero is True:
            attrs.append("H")
        elif self._hetero is False:
            attrs.append("C")

        return "n{}Ring".format("".join(attrs))

    def parameters(self):
        return self._order, self._greater, self._fused, self._aromatic, self._hetero

    def __init__(
        self, order=None, greater=False, fused=False, aromatic=None, hetero=None
    ):
        self._order = order
        self._greater = greater
        self._fused = fused
        self._aromatic = aromatic
        self._hetero = hetero

    def dependencies(self):
        return {"Rs": (FusedRings if self._fused else Rings)()}

    def _check_order(self, R):
        if self._order is None:
            return True

        if self._greater:
            return len(R) >= self._order
        else:
            return len(R) == self._order

    def _check_arom(self, R):
        if self._aromatic is None:
            return True

        is_arom = all(self.mol.GetAtomWithIdx(i).GetIsAromatic() for i in R)

        if self._aromatic:
            return is_arom

        return not is_arom

    def _check_hetero(self, R):
        if self._hetero is None:
            return True

        has_hetero = any(self.mol.GetAtomWithIdx(i).GetAtomicNum() != 6 for i in R)

        if self._hetero:
            return has_hetero

        return not has_hetero

    def calculate(self, Rs):
        return sum(
            1
            for R in Rs
            if self._check_order(R) and self._check_arom(R) and self._check_hetero(R)
        )

    rtype = int



# flake8: noqa: S1
# -*- coding: UTF-8 -*-

from __future__ import unicode_literals

import math

from rdkit import Chem

from ._base import Descriptor
from .Weight import Weight

__all__ = ("LogS",)


_smarts_logs = {
    "[NH0;X3;v3]": 0.71535,
    "[NH2;X3;v3]": 0.41056,
    "[nH0;X3]": 0.82535,
    "[OH0;X2;v2]": 0.31464,
    "[OH0;X1;v2]": 0.14787,
    "[OH1;X2;v2]": 0.62998,
    "[CH2;!R]": -0.35634,
    "[CH3;!R]": -0.33888,
    "[CH0;R]": -0.21912,
    "[CH2;R]": -0.23057,
    "[ch0]": -0.37570,
    "[ch1]": -0.22435,
    "F": -0.21728,
    "Cl": -0.49721,
    "Br": -0.57982,
    "I": -0.51547,
}


_smarts_logs_molecules = [
    (Chem.MolFromSmarts(smarts), log) for smarts, log in _smarts_logs.items()
]


class LogS(Descriptor):
    r"""Filter-it™ LogS descriptor.

    http://silicos-it.be.s3-website-eu-west-1.amazonaws.com/software/filter-it/1.0.2/filter-it.html#installation
    """

    __slots__ = ()

    since = "1.1.0"
    explicit_hydrogens = False
    kekulize = False

    @classmethod
    def preset(cls, version):
        yield cls()

    def dependencies(self):
        return {"MW": Weight(exact=False)}

    def description(self):
        return "Filter-it™ LogS"

    def __str__(self):
        return "FilterItLogS"

    def parameters(self):
        return ()

    def calculate(self, MW):
        logS = 0.89823 - 0.10369 * math.sqrt(MW)
        for smarts, log in _smarts_logs_molecules:
            logS += len(self.mol.GetSubstructMatches(smarts)) * log

        return logS

    rtype = float



import inspect

from mordred import descriptors, get_descriptors_in_module
from mordred._base.descriptor import is_descriptor_class


def test_attributes():
    for mdl in descriptors.all:
        for desc in get_descriptors_in_module(mdl, submodule=False):
            for cls in desc.__mro__:
                if cls == object:
                    continue

                if is_descriptor_class(desc, include_abstract=True):
                    assert (
                        "__slots__" in cls.__dict__
                    ), "{}({}) class don't have __slots__".format(
                        cls.__name__, inspect.getfile(desc)
                    )

            assert "since" in desc.__dict__, "{}({}) class don't have since".format(
                desc.__name__, inspect.getfile(desc)
            )



from mordred import (
    Calculator,
    descriptors,
)


def test_descriptor_order():
    calc = Calculator(descriptors)
    it = iter(calc.descriptors)
    before = next(it).__module__
    for current in it:
        current = current.__module__
        assert before <= current, "{!r} > {!r}".format(before, current)

        before = current



from rdkit.Chem import rdMolDescriptors

from ._base import Descriptor

__all__ = ("HBondAcceptor", "HBondDonor")


class HBondBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False

    @classmethod
    def preset(cls, version):
        yield cls()

    rtype = int


class HBondAcceptor(HBondBase):
    r"""hydrogen bond acceptor descriptor(rdkit wrapper)."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "number of hydrogen bond acceptor"

    def __str__(self):
        return "nHBAcc"

    def parameters(self):
        return ()

    def calculate(self):
        return rdMolDescriptors.CalcNumHBA(self.mol)


class HBondDonor(HBondBase):
    r"""hydrogen bond donor descriptor(rdkit wrapper)."""

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "number of hydrogen bond donor"

    def __str__(self):
        return "nHBDon"

    def parameters(self):
        return ()

    def calculate(self):
        return rdMolDescriptors.CalcNumHBD(self.mol)



from packaging.version import Version as StrictVersion

from ._base import Descriptor
from ._graph_matrix import AdjacencyMatrix as A
from ._matrix_attributes import SM1, methods, get_method

__all__ = ("AdjacencyMatrix",)

try:
    from importlib.metadata import version
except ImportError:
    from importlib_metadata import version

_version_remove_SM1_A = StrictVersion("1.1.0")


class AdjacencyMatrix(Descriptor):
    r"""adjacency matrix descriptor.

    :type type: :py:class:`str`
    :param type: :ref:`matrix_aggregating_methods`
    """

    since = "1.0.0"
    __slots__ = ("_type",)
    explicit_hydrogens = False

    def description(self):
        return "{} of adjacency matrix".format(self._type.__name__)

    @classmethod
    def preset(cls, version):
        if version >= _version_remove_SM1_A:
            return (cls(m) for m in methods if m != SM1)
        else:
            return map(cls, methods)

    def __str__(self):
        return "{}_A".format(self._type.__name__)

    def parameters(self):
        return (self._type,)

    def __init__(self, type="SpMax"):
        self._type = get_method(type)

    def dependencies(self):
        return {
            "result": self._type(
                A(self.explicit_hydrogens), self.explicit_hydrogens, self.kekulize
            )
        }

    def calculate(self, result):
        return result

    rtype = float



from __future__ import division

from ._base import Descriptor

__all__ = ("FragmentComplexity",)


class FragmentComplexity(Descriptor):
    r"""fragment complexity descriptor.

    .. math::
        {\rm fragCpx} = \left| B^2 - A^2 + A \right| + \frac{H}{100}

    where
    :math:`A` is number of atoms,
    :math:`B` is number of bonds,
    :math:`H` is number of hetero atoms

    References
        * :doi:`10.1021/ci050521b`

    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "fragment complexity"

    @classmethod
    def preset(cls, version):
        yield cls()

    explicit_hydrogens = False

    def parameters(self):
        return ()

    def __str__(self):
        return "fragCpx"

    def calculate(self):
        A = self.mol.GetNumAtoms()
        B = self.mol.GetNumBonds()
        H = sum(1 for a in self.mol.GetAtoms() if a.GetAtomicNum() != 6)
        return abs(B ** 2 - A ** 2 + A) + H / 100

    rtype = float



r"""Extended Topochemical Atom(ETA) descriptor.

References
    * :doi:`10.1021/ci0342066`
    * :doi:`10.1016/j.jhazmat.2013.03.023`

"""
from __future__ import division

import numpy as np
from rdkit import Chem

from . import _atomic_property as ap
from ._base import Descriptor
from ._util import atoms_to_numpy
from .RingCount import RingCount

__all__ = (
    "EtaCoreCount",
    "EtaShapeIndex",
    "EtaVEMCount",
    "EtaCompositeIndex",
    "EtaFunctionalityIndex",
    "EtaBranchingIndex",
    "EtaDeltaAlpha",
    "EtaEpsilon",
    "EtaDeltaEpsilon",
    "EtaDeltaBeta",
    "EtaPsi",
    "EtaDeltaPsi",
)


class AlterMolecule(Descriptor):
    __slots__ = ("explicit_hydrogens", "_saturated")

    kekulize = True
    require_connected = True

    def parameters(self):
        return self.explicit_hydrogens, self._saturated

    def __init__(self, explicit_hydrogens, saturated=False):
        self._saturated = saturated
        self.explicit_hydrogens = explicit_hydrogens

    def __str__(self):
        b = "Saturated" if self._saturated else "Reference"
        H = "H" if self.explicit_hydrogens else ""

        return "{}Mol{}".format(b, H)

    def calculate(self):
        new = Chem.RWMol(Chem.Mol())
        ids = {}
        for a in self.mol.GetAtoms():
            if a.GetAtomicNum() == 1:
                continue

            if self._saturated:
                new_a = Chem.Atom(a.GetAtomicNum())
                new_a.SetFormalCharge(a.GetFormalCharge())

            else:
                new_a = Chem.Atom(6)

            ids[a.GetIdx()] = new.AddAtom(new_a)

        for bond in self.mol.GetBonds():
            ai = bond.GetBeginAtom()
            aj = bond.GetEndAtom()

            if not self._saturated and (ai.GetDegree() > 4 or aj.GetDegree() > 4):
                self.fail(ValueError("bond degree greater then 4"))

            i = ids.get(ai.GetIdx())
            j = ids.get(aj.GetIdx())

            if i is not None and j is not None:
                if self._saturated and (
                    ai.GetAtomicNum() != 6 or aj.GetAtomicNum() != 6
                ):
                    order = bond.GetBondType()
                else:
                    order = Chem.BondType.SINGLE

                new.AddBond(i, j, order)

        new = Chem.Mol(new)
        if Chem.SanitizeMol(new, catchErrors=True) != 0:
            typ = "saturated" if self._saturated else "referense"
            self.fail(ValueError("cannot sanitize {} mol".format(typ)))

        if self.explicit_hydrogens:
            new = Chem.AddHs(new)

        Chem.Kekulize(new)

        return new


class EtaBase(Descriptor):
    __slots__ = ()

    explicit_hydrogens = False
    kekulize = True
    require_connected = True

    rtype = float


class EtaCoreCount(EtaBase):
    r"""ETA core count descriptor.

    .. math::

        \alpha = \sum_{i = 1}^A \frac{Z_i - Z_i^v}{Z_i^v} \cdot \frac{1}{PN_i - 1}

    where :math:`Z_i` and :math:`Z_i^v` are number of total and valence electons,
    :math:`PN` is periodic number.

    :type averaged: bool
    :param averaged: averaged by number of heavy count

    :type reference: bool
    :param reference: use reference alkane
        (same graph structure, but all atoms are carbon and all bonds are single bond)

    :returns: reference and valence of any atoms > 4
    """

    since = "1.0.0"
    __slots__ = ("_averaged", "_reference")

    def description(self):
        return "{}ETA core count{}".format(
            "averaged " if self._averaged else "",
            " for reference graph" if self._reference else "",
        )

    @classmethod
    def preset(cls, version):
        return map(cls, [False, True])

    def __str__(self):
        suffix = "_R" if self._reference else ""
        ave = "A" if self._averaged else ""

        return "{}ETA_alpha{}".format(ave, suffix)

    def parameters(self):
        return self._averaged, self._reference

    def __init__(self, averaged=False, reference=False):
        self._averaged = averaged
        self._reference = reference

    def dependencies(self):
        if self._reference:
            return {"rmol": AlterMolecule(self.explicit_hydrogens)}

    def calculate(self, rmol=None):
        mol = rmol if self._reference else self.mol

        v = sum(ap.get_core_count(a) for a in mol.GetAtoms())
        if self._averaged:
            v /= mol.GetNumAtoms()

        return v


class EtaShapeIndex(EtaBase):
    r"""ETA shape index descriptor.

    .. math::
        {\rm shape}_t = \frac{\alpha_t}{\alpha}

    where :math:`\alpha_t` is p(alpha value only atoms which bond to 1 heavy atom),
    y(3), or x(4).

    :type type: str
    :param type: one of shape_types
    """

    since = "1.0.0"
    __slots__ = ("_type",)

    def description(self):
        return "ETA shape index (type: {})".format(self._type)

    shape_types = ("p", "y", "x")
    _type_to_degree = {"p": 1, "y": 3, "x": 4}

    @classmethod
    def preset(cls, version):
        return (cls(t) for t in cls.shape_types)

    def __str__(self):
        return "ETA_shape_{}".format(self._type)

    def parameters(self):
        return (self._type,)

    def __init__(self, type="p"):
        assert type in self.shape_types

        self._type = type

    def dependencies(self):
        return {"a": EtaCoreCount(False)}

    def calculate(self, a):
        d = self._type_to_degree[self._type]

        return (
            sum(ap.get_core_count(a) for a in self.mol.GetAtoms() if a.GetDegree() == d)
            / a
        )


class EtaVEMCount(EtaBase):
    r"""ETA VEM(valence electron mobile) count descriptor.

    .. math::
        \beta^{\rm s} = \frac{1}{2} \sum^A_{i=1} \beta^{\rm s}_i

        \beta^{\rm s}_i = \sum^A_{j = 1} x_{ij}\sigma_{ij}

        x_{ij} = \begin{cases}
            0.5  & \left( \left| \epsilon_i - \epsilon_j \right| \leq 0.3 \right) \\
            0.75 & \left( \left| \epsilon_i - \epsilon_j \right| > 0.3 \right)
        \end{cases}

        \epsilon_i = - \alpha_i + 0.3 Z^{\rm v}

    where :math:`\sigma_{ij}` is sigma bond count between i and j.

    .. math::
        \beta^{\rm ns\delta} = \sum^A_{i = 1} \beta^{\rm ns\delta}_i

    where :math:`\beta^{\rm ns\delta}_i` is
    0.5 if i-th atom is making resonance with an aromatic ring.

    .. math::
        \beta^{\rm ns} = \frac{1}{2} \sum^A_{i=1} \beta^{\rm ns}_i

        \beta^{\rm ns}_i = \sum^A_{j = 1} y_{ij}\pi_{ij} + \beta^{\rm ns\delta}_i

        y_{ij} = \begin{cases}
            2.0 & \left( {\rm ij\ is\ aromatic\ bond} \right) \\
            1.5 & \left( \left| \epsilon_i - \epsilon_j \right| > 0.3 \right) \\
            1.0 & \left( \left| \epsilon_i - \epsilon_j \right| \leq 0.3 \right)
        \end{cases}

    where :math:`\pi_{ij}` is pi bond count between i and j.

    .. math::
        \beta = \beta^{\rm s} + \beta^{\rm ns}

    :type type: str
    :param type: one of beta_types

    :type averaged: bool
    :param averaged: averaged by heavy atom count
    """

    since = "1.0.0"
    __slots__ = ("_type", "_averaged")

    def description(self):
        if self._type == "":
            d = ""
        elif self._type == "s":
            d = "sigma contribution to "
        elif self._type == "ns":
            d = "nonsigma contribution to "
        else:
            d = "delta contribution to "

        return "{}{}valence electron mobile count".format(
            "averaged " if self._averaged else "", d
        )

    @classmethod
    def preset(cls, version):
        return (cls(b, a) for b in cls.beta_types for a in [False, True])

    def __str__(self):
        typ = "_{}".format(self._type) if self._type else ""
        ave = "A" if self._averaged else ""

        return "{}ETA_beta{}".format(ave, typ)

    beta_types = ("", "s", "ns", "ns_d")

    def parameters(self):
        return self._type, self._averaged

    def __init__(self, type="", averaged=False):
        assert type in self.beta_types
        self._type = type
        self._averaged = averaged

    def _get_beta_s(self, atom):
        return ap.get_eta_beta_sigma(atom) / 2.0

    def _get_beta_ns_d(self, atom):
        v = ap.get_eta_beta_delta(atom)
        return v

    def _get_beta_ns(self, atom):
        return ap.get_eta_beta_non_sigma(atom) / 2.0 + self._get_beta_ns_d(atom)

    def _get_beta_(self, atom):
        return self._get_beta_s(atom) + self._get_beta_ns(atom)

    def calculate(self):
        getter = getattr(self, "_get_beta_" + self._type)

        if getter:
            v = sum(getter(a) for a in self.mol.GetAtoms())

        if self._averaged:
            v /= self.mol.GetNumAtoms()

        return v


class EtaCompositeIndex(EtaBase):
    r"""ETA composite index descriptor.

    .. math::
        \eta = \sum_{i < j} \left( \frac{\gamma_i \gamma_j}{r_{ij}^2} \right)^{0.5}

        \gamma_i = \frac{\alpha_i}{\beta_i}

    where :math:`r_{ij}` is graph distance.

    .. math::
        \eta^{\rm local} = \sum_{i < j, r_{ij} = 1} \left( \gamma_i \gamma_j \right)^{0.5}

    :type reference: bool
    :param reference: use reference alkane.

    :type local: bool
    :param local: use :math:`\eta^{\rm local}`

    :type averaged: bool
    :param averaged: averaged

    :returns: reference and valence of any atoms > 4
    """

    since = "1.0.0"
    __slots__ = ("_reference", "_local", "_averaged")

    def description(self):
        return "{}{}ETA composite index{}".format(
            "averaged " if self._averaged else "",
            "local " if self._local else "",
            " for reference graph",
        )

    @classmethod
    def preset(cls, version):
        ft = [False, True]
        return (cls(r, l, a) for r in ft for l in ft for a in ft)

    def __str__(self):
        suffixes = []

        if self._reference:
            suffixes.append("R")

        if self._local:
            suffixes.append("L")

        if len(suffixes) > 0:
            suffix = "_" + "".join(suffixes)
        else:
            suffix = ""

        ave = "A" if self._averaged else ""

        return "{}ETA_eta{}".format(ave, suffix)

    def parameters(self):
        return self._reference, self._local, self._averaged

    def __init__(self, reference=False, local=False, averaged=False):
        self._reference = reference
        self._local = local
        self._averaged = averaged

    def dependencies(self):
        if self._reference:
            return {"rmol": AlterMolecule(self.explicit_hydrogens)}

    def calculate(self, rmol=None):
        mol = rmol if self._reference else self.mol
        D = Chem.GetDistanceMatrix(mol, force=True)

        if self._local:

            def checker(r):
                return r == 1

        else:

            def checker(r):
                return r != 0

        gamma = atoms_to_numpy(ap.get_eta_gamma, mol)

        v = sum(
            sum(
                np.sqrt(gamma[i] * gamma[j] / r ** 2)
                for j, r in enumerate(row)
                if i < j and checker(r)
            )
            for i, row in enumerate(D)
        )

        if self._averaged:
            v /= mol.GetNumAtoms()

        return v


class EtaFunctionalityIndex(EtaBase):
    r"""ETA functionality index descriptor.

    .. math::
        \eta^{\rm F} = \eta^{\rm R} - \eta

    where :math:`\eta^{\rm R}` is eta value for reference alkane.

    :type local: bool
    :param local: use local eta

    :type averaged: bool
    :param averaged: averaged
    """

    since = "1.0.0"
    __slots__ = ("_local", "_averaged")

    def description(self):
        return "{}{}ETA functionality index".format(
            "averaged " if self._averaged else "", "local " if self._local else ""
        )

    @classmethod
    def preset(cls, version):
        return (cls(l, a) for l in [False, True] for a in [False, True])

    def __str__(self):
        loc = "L" if self._local else ""
        ave = "A" if self._averaged else ""

        return "{}ETA_eta_F{}".format(ave, loc)

    def parameters(self):
        return self._local, self._averaged

    def __init__(self, local=False, averaged=False):
        self._local = local
        self._averaged = averaged

    def dependencies(self):
        return {
            "eta": EtaCompositeIndex(local=self._local),
            "eta_R": EtaCompositeIndex(local=self._local, reference=True),
        }

    def calculate(self, eta, eta_R):
        v = eta_R - eta
        if self._averaged:
            v /= self.mol.GetNumAtoms()

        return v


class EtaBranchingIndex(EtaBase):
    r"""ETA branching index descriptor.

    .. math::
        \eta^{\rm B} = \eta^{\rm local,N} - \eta^{local,R} + 0.086 N^{\rm R}

    where :math:`\eta^{\rm local,N}` is :math:`\eta^{\rm local}` for normal alkane.
    :math:`N^{\rm R}` is ring count.

    :type ring: bool
    :param ring: use ring count or not

    :type averaged: bool
    :param averaged: averaged

    :returns: NaN when A < 2
    """

    since = "1.0.0"
    __slots__ = ("_ring", "_averaged")

    def description(self):
        return "{}ETA branching index{}".format(
            "averaged " if self._averaged else "",
            " (use ring count)" if self._ring else "",
        )

    @classmethod
    def preset(cls, version):
        return (cls(r, a) for r in [False, True] for a in [False, True])

    def __str__(self):
        ring = "R" if self._ring else ""
        ave = "A" if self._averaged else ""

        return "{}ETA_eta_B{}".format(ave, ring)

    def parameters(self):
        return self._ring, self._averaged

    def __init__(self, ring=True, averaged=False):
        self._ring = ring
        self._averaged = averaged

    def dependencies(self):
        return {
            "NR": RingCount() if self._ring else None,
            "eta_RL": EtaCompositeIndex(reference=True, local=True),
        }

    def calculate(self, eta_RL, NR):
        N = self.mol.GetNumAtoms()

        if N <= 1:
            self.fail(ValueError("single atom"))
        elif N == 2:
            eta_NL = 1.0
        else:
            eta_NL = np.sqrt(2) + 0.5 * (N - 3)

        v = eta_NL - eta_RL + 0.086 * (NR or 0)

        if self._averaged:
            v /= N

        return v


class EtaDeltaAlpha(EtaBase):
    r"""ETA delta alpha descriptor.

    .. math::
        \Delta\alpha_{\rm A} = \max\left(\frac{\alpha - \alpha^{\rm R}}{A}, 0\right)

        \Delta\alpha_{\rm B} = \max\left(\frac{\alpha^{\rm R} - \alpha}{A}, 0\right)

    :type type: str
    :param type: one of delta_types
    """

    since = "1.0.0"
    __slots__ = ("_type",)

    def description(self):
        return "ETA delta alpha (type: {})".format(self._type)

    delta_types = ("A", "B")

    @classmethod
    def preset(cls, version):
        return (cls(t) for t in cls.delta_types)

    def __str__(self):
        return "ETA_dAlpha_{}".format(self._type)

    def parameters(self):
        return (self._type,)

    def __init__(self, type="A"):
        assert type in self.delta_types

        self._type = type

    def dependencies(self):
        return {"alpha": EtaCoreCount(), "alpha_R": EtaCoreCount(reference=True)}

    def calculate(self, alpha, alpha_R):
        if self._type == "A":
            d = alpha - alpha_R
        else:
            d = alpha_R - alpha

        return max(d / self.mol.GetNumAtoms(), 0.0)


class EtaEpsilon(EtaBase):
    r"""ETA epsilon descriptor.

    .. math::
        \epsilon^i = \frac{\epsilon^i}{N^i} (i \leq 4)
        \epsilon^5 = \frac{\epsilon^2 + \epsilon^{\rm XH}}{N^2 + N^{\rm XH}}

    types(i)
        1
            all atoms
        2
            heavy atoms
        3
            all atoms of reference alkane
        4
            all atoms of saturated carbon skeleton(reduce C-C bonds)
        XH
            hydrogens bond to hetero atoms

    :type type: str
    :param type: one of epsilon_types

    :returns: type = 3 and valence of any atoms > 4
    """

    since = "1.0.0"
    __slots__ = ("_type",)

    def description(self):
        return "ETA epsilon (type: {})".format(self._type)

    @classmethod
    def preset(cls, version):
        return map(cls, cls.epsilon_types)

    def __str__(self):
        return "ETA_epsilon_{}".format(self._type)

    @property
    def explicit_hydrogens(self):
        return self._type != 2

    epsilon_types = tuple(range(1, 6))

    def parameters(self):
        return (self._type,)

    def __init__(self, type=1):
        self._type = type

    def dependencies(self):
        if self._type == 3:
            return {"rmol": AlterMolecule(self.explicit_hydrogens)}
        elif self._type == 4:
            return {"rmol": AlterMolecule(self.explicit_hydrogens, True)}

    def calculate(self, rmol=None):
        mol = rmol if self._type in [3, 4] else self.mol

        if self._type == 5:
            eps = [
                ap.get_eta_epsilon(a)
                for a in mol.GetAtoms()
                if a.GetAtomicNum() != 1 or a.GetNeighbors()[0].GetAtomicNum() != 6
            ]
            return sum(eps) / len(eps)

        return sum(ap.get_eta_epsilon(a) for a in mol.GetAtoms()) / mol.GetNumAtoms()


class EtaDeltaEpsilon(EtaBase):
    r"""ETA delta epsilon descriptor.

    .. math::
        \Delta \epsilon^{\rm A} = \epsilon^1 - \epsilon^3

        \Delta \epsilon^{\rm B} = \epsilon^1 - \epsilon^4

        \Delta \epsilon^{\rm C} = \epsilon^3 - \epsilon^4

        \Delta \epsilon^{\rm D} = \epsilon^2 - \epsilon^5

    :type type: str
    :param type: one of delta_epsilon_types
    """

    since = "1.0.0"
    __slots__ = ("_type",)

    def description(self):
        return "ETA delta epsilon (type: {})".format(self._type)

    @classmethod
    def preset(cls, version):
        return map(cls, cls.delta_epsilon_types)

    def __str__(self):
        return "ETA_dEpsilon_{}".format(self._type)

    delta_epsilon_types = tuple("ABCD")

    def parameters(self):
        return (self._type,)

    def __init__(self, type="A"):
        self._type = type

    _deps = {"A": (1, 3), "B": (1, 4), "C": (3, 4), "D": (2, 5)}

    def dependencies(self):
        L, R = self._deps[self._type]
        return {"L": EtaEpsilon(L), "R": EtaEpsilon(R)}

    def calculate(self, L, R):
        return L - R


class EtaDeltaBeta(EtaBase):
    r"""ETA delta beta descriptor.

    .. math::
        \Delta\beta = \beta^{\rm ns} - \beta^{\rm s}

    :type averaged: bool
    :param averaged: averaged
    """

    since = "1.0.0"
    __slots__ = ("_averaged",)

    def description(self):
        return "{}ETA delta beta".format("averaged " if self._averaged else "")

    @classmethod
    def preset(cls, version):
        return (cls(a) for a in [False, True])

    def __str__(self):
        ave = "A" if self._averaged else ""

        return "{}ETA_dBeta".format(ave)

    def parameters(self):
        return (self._averaged,)

    def __init__(self, averaged=False):
        self._averaged = averaged

    def dependencies(self):
        return {"ns": EtaVEMCount("ns"), "s": EtaVEMCount("s")}

    def calculate(self, ns, s):
        v = ns - s

        if self._averaged:
            v /= self.mol.GetNumAtoms()

        return v


class EtaPsi(EtaBase):
    r"""ETA psi descriptor.

    .. math::
        \psi_1 = \frac{\alpha}{A \cdot \epsilon^2}
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "ETA psi"

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return "ETA_psi_1"

    def parameters(self):
        return ()

    def dependencies(self):
        return {"a": EtaCoreCount(), "e": EtaEpsilon(2)}

    def calculate(self, a, e):
        return a / (self.mol.GetNumAtoms() * e)


class EtaDeltaPsi(EtaBase):
    r"""ETA delta psi descriptor.

    .. math::
        \Delta\psi_{\rm A} = \max\left(0.714 - \psi_1, 0\right)

        \Delta\psi_{\rm B} = \max\left(\psi_1 - 0.714, 0\right)

    :type type: str
    :param type: one of delta_psi_types
    """

    since = "1.0.0"
    __slots__ = ("_type",)

    def description(self):
        return "ETA delta psi (type: {})".format(self._type)

    @classmethod
    def preset(cls, version):
        return map(cls, cls.delta_psi_types)

    def __str__(self):
        return "ETA_dPsi_{}".format(self._type)

    delta_psi_types = ("A", "B")

    def parameters(self):
        return (self._type,)

    def __init__(self, type="A"):
        assert type in self.delta_psi_types

        self._type = type

    def dependencies(self):
        return {"psi": EtaPsi()}

    def calculate(self, psi):
        L = 0.714
        R = psi

        if self._type == "B":
            L, R = R, L

        return max(L - R, 0.0)



from itertools import islice
from collections import deque
from multiprocessing import Pool, Manager

from .._util import Capture
from .context import Context

calculator = None


def worker(calc_proxy, cxt):
    global calculator
    if calculator is None:
        calculator = calc_proxy[0]

    with Capture() as capture:
        r = list(calculator._calculate(cxt))

        return r, capture.result


class MolPool(object):
    def __init__(self, calc, nproc):
        self.pool = Pool(nproc)
        self.mgr = Manager()
        self.calc = calc
        self.calc_proxy = self.mgr.list([calc])
        self.nproc = nproc

    def __enter__(self):
        self.mgr.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self.pool.terminate()
        self.mgr.__exit__(*args, **kwargs)

    def map(self, mols, id):
        return MolIterator(self, mols, id, self.nproc * 2 + 10)

    def submit(self, mol, id):
        cxt = Context.from_calculator(self.calc, mol, id)
        return self.pool.apply_async(worker, (self.calc_proxy, cxt))


class MolIterator(object):
    def __init__(self, pool, mols, id, buf):
        self.pool = pool
        self.futures = deque()
        self.mols = iter(mols)
        self.id = id

        for mol in islice(self.mols, buf):
            self.submit(mol)

    def submit(self, mol):
        self.futures.append((mol, self.pool.submit(mol, self.id)))

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self.submit(next(self.mols))
        except StopIteration:
            pass

        try:
            mol, fut = self.futures.popleft()
            return mol, fut.get()
        except IndexError:
            raise StopIteration

    next = __next__


def parallel(calc, mols, nproc, nmols, quiet, ipynb, id):
    with MolPool(calc, nproc) as pool, calc._progress(quiet, nmols, ipynb) as bar:
        for mol, (r, err) in pool.map(mols, id):
            for e in err:
                e = e.rstrip()
                if not e:
                    continue

                bar.write(e)

            yield calc._wrap_result(mol, r)
            bar.update()



import os

from rdkit.Chem import AllChem as Chem
from numpy.testing import assert_almost_equal

from mordred import Calculator, descriptors
from mordred.error import MissingValueBase

data_file = os.path.join(os.path.dirname(__file__), "references", "structures.sdf")


def test_parallel():
    calc = Calculator(descriptors)
    mols = [m for m in Chem.SDMolSupplier(data_file, removeHs=False)]

    for serial, parallel in zip(
        calc.map(mols, nproc=1, quiet=True), calc.map(mols, quiet=True)
    ):
        for d, s, p in zip(calc.descriptors, serial, parallel):
            if isinstance(s, MissingValueBase):
                assert s.error.__class__ == p.error.__class__
            else:
                msg = "{} (serial: {}, parallel: {})".format(str(d), s, p)
                assert assert_almost_equal(s, p, 7, msg) is None



from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred.ABCIndex import ABCIndex, ABCGGIndex

# doi:10.2298/JSC150901093F
data = """
CC(C)CCCCCCC   6.58 6.49
CCC(C)CCCCCC   6.47 6.58
CC(C)(C)CCCCCC 6.84 6.82
CCC(C)(C)CCCCC 6.68 6.95
"""[
    1:-1
]


def test_ABC():
    abc = ABCIndex()
    abcgg = ABCGGIndex()

    for line in data.split("\n"):
        smi, dABC, dABCGG = line.strip().split()

        mol = Chem.MolFromSmiles(smi)
        assert assert_almost_equal(abc(mol), float(dABC), 2) is None
        assert assert_almost_equal(abcgg(mol), float(dABCGG), 2) is None



from math import pi

from ._base import Descriptor
from .BondCount import BondCount
from .RingCount import RingCount
from ._atomic_property import GetAtomicNumber

__all__ = ("VdwVolumeABC",)


bondi_radii = {
    "H": 1.20,
    "C": 1.70,  # noqa: S001
    "N": 1.55,
    "O": 1.52,
    "F": 1.47,  # noqa: S001
    "Cl": 1.75,  # noqa: S001
    "Br": 1.85,  # noqa: S001
    "P": 1.80,
    "S": 1.80,
    "As": 1.85,  # noqa: S001
    "B": 2.13,
    "Si": 2.10,
    "Se": 1.90,  # noqa: S001
}


atom_contrib = {
    GetAtomicNumber(s): 4.0 / 3.0 * pi * r ** 3 for s, r in bondi_radii.items()
}


class VdwVolumeABC(Descriptor):
    r"""van der waals volume(ABC) descriptor.

    :returns: NaN when any atoms are non-compat_atoms

    References
        * :doi:`10.1021/jo034808o`

    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "ABC van der waals volume"

    compat_atoms = tuple(bondi_radii)

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return "Vabc"

    def parameters(self):
        return ()

    def dependencies(self):
        return {
            "NRA": RingCount(None, False, False, False, None),
            "NRa": RingCount(None, False, False, True, None),
            "Nb": BondCount(),
        }

    def calculate(self, Nb, NRa, NRA):
        try:
            ac = sum(atom_contrib[a.GetAtomicNum()] for a in self.mol.GetAtoms())
        except KeyError:
            self.fail(ValueError("unknown atom type"))

        return ac - 5.92 * Nb - 14.7 * NRa - 3.8 * NRA

    rtype = float



r"""modred descriptor calculator."""

from ._base import (
    Result,
    Calculator,
    Descriptor,
    is_missing,
    get_descriptors_in_module,
)

__all__ = (
    "Descriptor",
    "Calculator",
    "get_descriptors_in_module",
    "is_missing",
    "Result",
)



import numpy as np
from pandas import Series, DataFrame

from .util import is_missing

__all__ = ("MordredDataFrame", "Series")


class MordredDataFrame(DataFrame):
    @property
    def _constructor(self):
        return self.__class__

    def fill_missing(self, value=np.nan, inplace=False):
        t = self if inplace else self.copy()

        t[t.applymap(is_missing)] = value
        return t



from ._base import Descriptor
from ._graph_matrix import Valence, Eccentricity

__all__ = ("EccentricConnectivityIndex",)


class EccentricConnectivityIndex(Descriptor):
    r"""eccentric connectivity index descriptor.

    .. math::
        I_{\rm EC} = \sum^A_i {\boldsymbol E}{\boldsymbol V}

    where
    :math:`E` is eccentricity of atoms,
    :math:`V` is valences of atoms.
    """

    since = "1.0.0"
    __slots__ = ()
    explicit_hydrogens = False

    def description(self):
        return "eccentric connectivity index"

    @classmethod
    def preset(cls, version):
        yield cls()

    def __str__(self):
        return "ECIndex"

    def parameters(self):
        return ()

    def dependencies(self):
        return {
            "E": Eccentricity(self.explicit_hydrogens),
            "V": Valence(self.explicit_hydrogens),
        }

    def calculate(self, E, V):
        return int((E.astype("int") * V).sum())

    rtype = int



import time

import numpy as np
import networkx

from . import _matrix_attributes as ma
from ._base import Descriptor
from .error import Timeout

__all__ = ("DetourMatrix", "DetourIndex")


class LongestSimplePath(object):
    __slots__ = (
        "G",
        "N",
        "neighbors",
        "start",
        "result",
        "visited",
        "distance",
        "timeout_at",
    )

    def __init__(self, G, weight=None, timeout_at=None):
        self.G = G
        self.N = G.number_of_nodes()
        self.timeout_at = timeout_at
        self.neighbors = {
            n: [(v, d.get(weight, 1.0)) for v, d in G[n].items()] for n in G.nodes()
        }

    def _start(self, s):
        self.start = s
        self.result = {n: 0 for n in self.G.nodes()}
        self.visited = set()
        self.distance = 0.0
        self._search(s)
        return self.result

    def _search(self, u):
        if self.timeout_at < time.time():
            raise Timeout()

        self.visited.add(u)
        for v, w in self.neighbors[u]:
            if v in self.visited:
                continue

            self.visited.add(v)
            self.distance += w

            d = self.distance
            if d > self.result[v]:
                self.result[v] = d

            self._search(v)

            self.visited.remove(v)
            self.distance -= w

    def __call__(self):
        return {
            (min(s, g), max(s, g)): w
            for s in self.G.nodes()
            for g, w in self._start(s).items()
        }


class CalcDetour(object):
    __slots__ = ("N", "G", "Q", "nodes", "C", "weight", "timeout")

    def __init__(self, G, weight="weight", timeout=None):
        self.G = G
        self.N = G.number_of_nodes()
        self.Q = []
        self.weight = weight
        self.timeout = timeout

    def merge(self):
        for i in range(1, len(self.Q) + 1):
            ns, lsp = self.Q[-i]
            common = ns & self.nodes
            if len(common) == 0:
                continue
            elif len(common) > 1:
                raise ValueError("bug: multiple common nodes.")

            common = common.pop()
            self.Q.pop(-i)
            for n in ns:
                self.nodes.add(n)
            break

        def calc_weight(i, j):
            ij = (i, j)
            if ij in self.C:
                return self.C[ij]
            elif ij in lsp:
                return lsp[ij]
            elif i == j == common:
                return max(lsp[ij], self.C[ij])

            ic = (min(i, common), max(i, common))
            jc = (min(j, common), max(j, common))

            if ic in self.C and jc in lsp:
                return self.C[ic] + lsp[jc]
            elif jc in self.C and ic in lsp:
                return self.C[jc] + lsp[ic]
            else:
                raise ValueError("bug: unknown weight")

        self.C = {
            (i, j): calc_weight(i, j) for i in self.nodes for j in self.nodes if i <= j
        }

    def __call__(self):
        timeout_at = None if self.timeout is None else time.time() + self.timeout

        for bcc in (
            self.G.subgraph(c) for c in networkx.biconnected_components(self.G)
        ):
            lsp = LongestSimplePath(bcc, self.weight, timeout_at)()
            nodes = set()
            for a, b in lsp:
                nodes.add(a)
                nodes.add(b)
            self.Q.append((nodes, lsp))

        if self.N == 1:
            return np.array([[0]])

        self.nodes, self.C = self.Q.pop()

        while self.Q:
            self.merge()

        result = np.empty((self.N, self.N))
        for i, j in ((i, j) for i in range(self.N) for j in range(i, self.N)):
            result[i, j] = self.C[(i, j)]
            result[j, i] = self.C[(i, j)]

        return result


class DetourMatrixBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False
    require_connected = True


class DetourMatrixCache(DetourMatrixBase):
    __slots__ = ()

    hermitian = True

    def parameters(self):
        return ()

    def calculate(self):
        G = networkx.Graph()
        G.add_nodes_from(a.GetIdx() for a in self.mol.GetAtoms())
        G.add_edges_from(
            (b.GetBeginAtomIdx(), b.GetEndAtomIdx()) for b in self.mol.GetBonds()
        )

        return CalcDetour(G, timeout=self.config.get("timeout", 60))()


class DetourMatrix(DetourMatrixBase):
    r"""detour matrix descriptor.

    :type type: str
    :param type: :ref:`matrix_aggregating_methods`
    """

    since = "1.0.0"
    __slots__ = ("_type",)

    def description(self):
        return "{} from detourn matrix".format(self._type.description())

    @classmethod
    def preset(cls, version):
        return map(cls, ma.methods)

    def __str__(self):
        return "{}_Dt".format(self._type.__name__)

    def parameters(self):
        return (self._type,)

    def __init__(self, type="SpMax"):
        self._type = ma.get_method(type)

    def dependencies(self):
        return {
            "result": self._type(
                DetourMatrixCache(), self.explicit_hydrogens, self.kekulize
            )
        }

    def calculate(self, result):
        return result

    rtype = float


class DetourIndex(DetourMatrixBase):
    r"""detour index descriptor.

    .. math::

        I_{\rm D} = \frac{1}{A} \sum^A_{i=1} \sum^A_{j=1} {\boldsymbol D}_{ij}

    where
    :math:`D` is detour matrix,
    :math:`A` is number of atoms.
    """

    since = "1.0.0"
    __slots__ = ()

    def description(self):
        return "detour index"

    def parameters(self):
        return ()

    @classmethod
    def preset(cls, version):
        yield cls()

    explicit_hydrogens = False

    def __str__(self):
        return self.__class__.__name__

    def dependencies(self):
        return {"D": DetourMatrixCache()}

    def calculate(self, D):
        return int(0.5 * D.sum())

    rtype = int



from collections import defaultdict

from rdkit.Chem import BondType, rdMolDescriptors

from ._base import Descriptor

__all__ = ("TopoPSA",)


class TopoPSA(Descriptor):
    r"""topological polar surface area descriptor(NO only: rdkit wrapper).

    :type no_only: bool
    :param no_only:
        * True: N,O only TPSA
        * False: all(N,O,S,P) TPSA

    References
        * :doi:`10.1021/jm000942e`

    """

    since = "1.0.0"
    __slots__ = ("_no_only",)

    def description(self):
        return "topological polar surface area{}".format(
            " (use only nitrogen and oxygen)" if self._no_only else ""
        )

    @classmethod
    def preset(cls, version):
        yield cls(True)
        yield cls(False)

    def __str__(self):
        return "TopoPSA(NO)" if self._no_only else "TopoPSA"

    def parameters(self):
        return (self._no_only,)

    def __init__(self, no_only=True):
        self._no_only = no_only

    def calculate(self):
        tpsa = rdMolDescriptors.CalcTPSA(self.mol)
        if self._no_only:
            return tpsa

        for atom in self.mol.GetAtoms():
            atomic_num = atom.GetAtomicNum()

            if atomic_num == 15:
                tpsa += self._get_phosphorus_contrib(atom)
            elif atomic_num == 16:
                tpsa += self._get_sulfur_contrib(atom)

        return tpsa

    @staticmethod
    def _hydrogen_count(atom):
        return atom.GetTotalNumHs() + sum(
            1 for a in atom.GetNeighbors() if a.GetAtomicNum() == 1
        )

    @staticmethod
    def _bond_type_count(atom):
        cnt = defaultdict(int)
        for bond in atom.GetBonds():
            if bond.GetIsAromatic():
                cnt[BondType.AROMATIC] += 1
            else:
                cnt[bond.GetBondType()] += 1

        return dict(cnt)

    @classmethod
    def _get_phosphorus_contrib(cls, atom):
        nH = cls._hydrogen_count(atom)
        cnt = cls._bond_type_count(atom)

        if atom.GetFormalCharge() != 0 or atom.GetIsAromatic():
            return 0.0

        if nH == 1 and cnt == {BondType.SINGLE: 3, BondType.DOUBLE: 1}:
            return 23.47
        elif nH == 0:
            if cnt == {BondType.SINGLE: 3}:
                return 13.59
            elif cnt == {BondType.SINGLE: 1, BondType.DOUBLE: 1}:
                return 34.14
            elif cnt == {BondType.SINGLE: 3, BondType.DOUBLE: 1}:
                return 9.81

        return 0.0

    @classmethod
    def _get_sulfur_contrib(cls, atom):
        nH = cls._hydrogen_count(atom)
        cnt = cls._bond_type_count(atom)

        if atom.GetFormalCharge() != 0:
            return 0.0

        if atom.GetIsAromatic():
            if nH == 0:
                if cnt == {BondType.AROMATIC: 2}:
                    return 28.24
                elif cnt == {BondType.AROMATIC: 2, BondType.DOUBLE: 1}:
                    return 21.70

        else:
            if nH == 1 and cnt == {BondType.SINGLE: 2}:
                return 38.80
            elif nH == 0:
                if cnt == {BondType.SINGLE: 2}:
                    return 25.30
                elif cnt == {BondType.DOUBLE: 1}:
                    return 32.09
                elif cnt == {BondType.SINGLE: 2, BondType.DOUBLE: 1}:
                    return 19.21
                elif cnt == {BondType.SINGLE: 2, BondType.DOUBLE: 2}:
                    return 8.38

        return 0.0

    rtype = float



try:
    from collections.abc import Iterable
except ImportError:
    import warnings

    warnings.filterwarnings(
        "ignore",
        ".* ABCs from 'collections' .*",
        DeprecationWarning,
    )
    from collections import Iterable

import numpy as np

from mordred import Result, Descriptor, error, is_missing


class Dummy1(Descriptor):
    def __str__(self):
        return "Dummy1"

    def parameters(self):
        return ()

    def calculate(self):
        return 1


class Dummy2(Descriptor):
    def __str__(self):
        return "Dummy2"

    def parameters(self):
        return ()

    def calculate(self):
        raise ValueError("error")


result = Result(None, [1, error.Error(ValueError("error"), [])], [Dummy1(), Dummy2()])


def test_length():
    assert len(result) == 2


def test_fill_missing():
    assert np.isnan(result.fill_missing()[1])


def test_drop_missing():
    assert len(result.drop_missing()) == 1


def test_items():
    i = result.items()
    assert isinstance(i, Iterable)
    li = list(i)
    assert len(li) == 2
    assert len(li[0]) == 2

    d, v = li[0]
    assert isinstance(d, Descriptor)
    assert isinstance(v, int)


def test_keys():
    i = result.keys()
    assert isinstance(i, Iterable)
    li = list(i)
    assert len(li), 2

    assert isinstance(li[0], Descriptor)


def test_iter():
    i = iter(result)
    assert isinstance(i, Iterable)
    li = list(i)
    assert len(li) == 2

    assert isinstance(li[0], int)


def test_reversed():
    i = reversed(result)
    assert isinstance(i, Iterable)
    li = list(i)
    assert len(li) == 2

    assert isinstance(li[1], int)


def test_asdict_non_rawkey():
    d = result.asdict()
    assert len(d) == 2
    assert isinstance(d, dict)
    for k in d.keys():
        assert isinstance(k, str)


def test_asdict_rawkey():
    d = result.asdict(True)
    assert len(d) == 2
    assert isinstance(d, dict)
    for k in d.keys():
        assert isinstance(k, Descriptor)


def test_ix():
    assert 1 == result.ix[0]
    assert is_missing(result.ix[1])


def test_name():
    assert 1 == result.name["Dummy1"]
    assert is_missing(result.name["Dummy2"])


def test_getitem():
    assert 1 == result[0]
    assert is_missing(result[1])
    assert 1 == result["Dummy1"]
    assert is_missing(result["Dummy2"])



from enum import IntEnum
from itertools import chain
from collections import namedtuple, defaultdict

from rdkit import Chem

from ._base import Descriptor
from ._util import parse_enum
from ._atomic_property import AtomicProperty

__all__ = ("Chi",)


class ChiType(IntEnum):
    __slots__ = ()

    path = 1
    cluster = 2
    path_cluster = 3
    chain = 4

    @property
    def as_argument(self):
        return self.name

    @property
    def short(self):
        _short_dict = {
            self.path: "p",
            self.chain: "ch",
            self.path_cluster: "pc",
            self.cluster: "c",
        }

        return _short_dict[self]

    @property
    def long(self):
        _long_dict = {
            self.path: "Chi path",
            self.chain: "Chi chain",
            self.path_cluster: "Chi path-cluster",
            self.cluster: "Chi cluster",
        }

        return _long_dict[self]


class DFS(object):
    __slots__ = (
        "mol",
        "visited",
        "vis_edges",
        "is_chain",
        "degrees",
        "bonds",
        "neighbors",
    )

    def __init__(self, mol):
        self.mol = mol
        self.visited = set()
        self.vis_edges = set()
        self.degrees = set()
        self.neighbors = defaultdict(set)

        self.bonds = [
            (b.GetBeginAtomIdx(), b.GetEndAtomIdx()) for b in self.mol.GetBonds()
        ]

    def reset(self, use_bonds):
        ns = self.neighbors
        bs = self.bonds

        self.is_chain = False
        self.visited.clear()
        self.vis_edges.clear()
        self.degrees.clear()
        ns.clear()

        for i in range(len(use_bonds)):
            a, b = bs[use_bonds[i]]
            ns[a].add(b)
            ns[b].add(a)

        self.neighbors = ns

    @property
    def nodes(self):
        return list(self.neighbors.keys())

    def _dfs(self, u):
        neighbors = self.neighbors[u]
        self.visited.add(u)
        self.degrees.add(len(neighbors))

        for v in neighbors:
            ek = (v, u) if u > v else (u, v)

            if v not in self.visited:
                self.vis_edges.add(ek)
                self._dfs(v)

            elif ek not in self.vis_edges:
                self.vis_edges.add(ek)
                self.is_chain = True

    def __call__(self):
        self._dfs(next(iter(self.neighbors.keys())))

        if self.is_chain:
            t = ChiType.chain
        elif not self.degrees - {1, 2}:
            t = ChiType.path
        elif 2 in self.degrees:
            t = ChiType.path_cluster
        else:
            t = ChiType.cluster

        return t


class ChiBase(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False


ChiBonds = namedtuple("ChiBonds", "chain path path_cluster cluster")


class ChiCache(ChiBase):
    __slots__ = ("_order",)

    def parameters(self):
        return (self._order,)

    def __init__(self, order):
        self._order = order

    def calculate(self):
        chain = []
        path = []
        path_cluster = []
        cluster = []

        dfs = DFS(self.mol)
        for bonds in Chem.FindAllSubgraphsOfLengthN(self.mol, self._order):
            dfs.reset(bonds)
            typ = dfs()
            nodes = dfs.nodes

            if typ == ChiType.chain:
                chain.append(nodes)
            elif typ == ChiType.path:
                path.append(nodes)
            elif typ == ChiType.path_cluster:
                path_cluster.append(nodes)
            else:
                cluster.append(nodes)

        return ChiBonds(chain, path, path_cluster, cluster)


class Chi(ChiBase):
    r"""chi descriptor.

    :type type: str
    :param type: one of chi_types

    :type prop: str or function
    :param prop: :ref:`atomic_properties`

    :type averaged: bool
    :param averaged: averaged by number of subgraphs

    :returns: NaN when

        * any atomic properties <= 0
        * averaged and :math:`N_{\chi} = 0`
    """

    since = "1.0.0"
    __slots__ = ("_type", "_order", "_prop", "_averaged")

    chi_types = tuple(t.name for t in ChiType)

    _deltas = ["d", "dv"]

    def description(self):
        return "{}-ordered {}{} weighted by {}".format(
            self._order,
            "averaged " if self._averaged else "",
            self._type.long,
            self._prop.get_long(),
        )

    @classmethod
    def preset(cls, version):
        return chain(
            (cls(ChiType.chain, l, a) for a in cls._deltas for l in range(3, 8)),
            (cls(ChiType.cluster, l, a) for a in cls._deltas for l in range(3, 7)),
            (cls(ChiType.path_cluster, l, a) for a in cls._deltas for l in range(4, 7)),
            (
                cls(ChiType.path, l, a, m)
                for a in cls._deltas
                for m in [False, True]
                for l in range(8)
            ),
        )

    def __str__(self):
        prop = self._prop.as_argument
        ct = self._type.short
        averaged = "A" if self._averaged else ""

        return "{}X{}-{}{}".format(averaged, ct, self._order, prop)

    def parameters(self):
        return self._type, self._order, self._prop, self._averaged

    def __init__(self, type="path", order=0, prop="d", averaged=False):
        self._type = parse_enum(ChiType, type)
        self._order = order
        self._prop = AtomicProperty(self.explicit_hydrogens, prop)
        self._averaged = averaged

    def dependencies(self):
        d = {"P": self._prop}
        if self._order > 0:
            d["chi"] = ChiCache(self._order)

        return d

    def calculate(self, P, chi=None):
        if self._order <= 0:
            chi = ChiBonds([], [{a.GetIdx()} for a in self.mol.GetAtoms()], [], [])

        x = 0.0
        node_sets = getattr(chi, self._type.name)
        for nodes in node_sets:
            c = 1
            for node in nodes:
                c *= P[node]

            if c <= 0:
                self.fail(ValueError("some properties less then or equal to 0"))

            x += c ** -0.5

        if self._averaged:
            with self.rethrow_zerodiv():
                x /= len(node_sets)

        return x

    rtype = float

    _extra_docs = ("chi_types",)



import math
import pickle

import six
from rdkit import Chem
from numpy.testing import assert_almost_equal

from mordred import Calculator, descriptors
from mordred.error import MissingValueBase


def test_pickle_calculator():
    orig = Calculator(descriptors)
    d0 = orig.descriptors[0]
    d1 = orig.descriptors[1]
    orig.register(
        [
            d0 + d1,
            d0 - d1,
            d0 * d1,
            d0 // d1,
            d0 % d1,
            d0**d1,
            -d0,
            +d1,
            abs(d0),
            math.trunc(d0),
        ]
    )

    if six.PY3:
        orig.register([math.ceil(d0), math.floor(d1)])

    pickled = pickle.loads(pickle.dumps(orig))

    mol = Chem.MolFromSmiles("c1ccccc1C(O)O")

    for a, b in zip(orig.descriptors, pickled.descriptors):
        assert a == b

    for a, b in zip(orig(mol), pickled(mol)):
        if isinstance(a, MissingValueBase):
            assert a.__class__ == b.__class__
        else:
            assert_almost_equal(a, b)



import numpy as np

from mordred.error import Error, Missing
from mordred._base.pandas_module import MordredDataFrame

df = MordredDataFrame(
    [
        [1, 2],
        [1, Error(ValueError("error"), ())],
        [Missing(ValueError("missing"), ()), 4],
    ]
)


def test_copy():
    assert isinstance(df.copy(), MordredDataFrame)


def test_fill_missing():
    f = df.fill_missing()
    assert f.iloc[0, 0], 1
    assert f.iloc[0, 1], 2
    assert f.iloc[1, 0], 1
    assert np.isnan(f.iloc[1, 1])
    assert np.isnan(f.iloc[2, 0])
    assert f.iloc[2, 1], 4



import json

from mordred import Calculator, descriptors


def test_json():
    calc = Calculator(descriptors)
    j = json.dumps(calc.to_json())
    calc2 = Calculator.from_json(json.loads(j))
    assert calc.descriptors == calc2.descriptors



import numpy as np

from ._base import Descriptor
from ._graph_matrix import AdjacencyMatrix

__all__ = ("WalkCount",)


class WalkCount(Descriptor):
    r"""walk count descriptor.

    :type order: int
    :param order: walk length

    :type total: bool
    :param total: sum of walk count over 1 to order

    :type self_returning: bool
    :param self_returning: use self returning walk only
    """

    since = "1.0.0"
    __slots__ = ("_order", "_total", "_self_returning")

    explicit_hydrogens = False

    def description(self):
        return "{}walk count (leg-{}{})".format(
            "total " if self._total else "",
            self._order,
            ", only self returning walk" if self._self_returning else "",
        )

    @classmethod
    def preset(cls, version):
        for start, sr in [(1, False), (2, True)]:
            for l in range(start, 11):
                yield cls(l, False, sr)

            yield cls(10, True, sr)

    def __str__(self):
        T = "{}SRW{:02d}" if self._self_returning else "{}MWC{:02d}"
        return T.format("T" if self._total else "", self._order)

    def parameters(self):
        return self._order, self._total, self._self_returning

    def __init__(self, order=1, total=False, self_returning=False):
        self._order = order
        self._total = total
        self._self_returning = self_returning

    def dependencies(self):
        if self._total:
            d = {}
            d["W"] = self.__class__(self._order, False, self._self_returning)

            if self._order > 1:
                d["T"] = self.__class__(self._order - 1, True, self._self_returning)

            return d

        return {"An": AdjacencyMatrix(self.explicit_hydrogens, order=self._order)}

    def calculate(self, An=None, T=None, W=None):
        if self._total:
            if self._order == 1:
                return self.mol.GetNumAtoms() + W

            return T + W

        if self._self_returning:
            return np.log(An.trace() + 1)

        else:
            if self._order == 1:
                return 0.5 * An.sum()

            return np.log(An.sum() + 1)

    rtype = float



from __future__ import print_function

import sys
import warnings
from types import ModuleType
from contextlib import contextmanager
from multiprocessing import cpu_count
from packaging.version import Version as StrictVersion

from .._util import Capture, DummyBar
from ..error import Error, Missing, MultipleFragments, DuplicatedDescriptorName
from .result import Result
from .context import Context
from .descriptor import Descriptor, MissingValueException, is_descriptor_class

try:
    from importlib.metadata import version as importlib_version
except ImportError:
    from importlib_metadata import version as importlib_version

__version__ = importlib_version("mordredcommunity")

try:
    from tqdm import tqdm
    from .._util import NotebookWrapper
except ImportError:
    tqdm = NotebookWrapper = DummyBar


class Calculator(object):
    r"""descriptor calculator.

    Parameters:
        descs: see Calculator.register() method
        ignore_3D: see Calculator.register() method

    """

    __slots__ = (
        "_descriptors",
        "_name_dict",
        "_explicit_hydrogens",
        "_kekulizes",
        "_require_3D",
        "_cache",
        "_debug",
        "_progress_bar",
        "_config",
    )

    def __setstate__(self, dict):
        ds = self._descriptors = dict.get("_descriptors", [])
        self._name_dict = {str(d): d for d in ds}
        self._explicit_hydrogens = dict.get("_explicit_hydrogens", {True, False})
        self._kekulizes = dict.get("_kekulizes", {True, False})
        self._require_3D = dict.get("_require_3D", False)

    @classmethod
    def from_json(cls, obj):
        """Create Calculator from json descriptor objects.

        Parameters:
            obj(list or dict): descriptors to register

        Returns:
            Calculator: calculator

        """
        calc = cls()
        calc.register_json(obj)
        return calc

    def register_json(self, obj):
        """Register Descriptors from json descriptor objects.

        Parameters:
            obj(list or dict): descriptors to register

        """
        if not isinstance(obj, list):
            obj = [obj]

        self.register(Descriptor.from_json(j) for j in obj)

    def to_json(self):
        """Convert descriptors to json serializable data.

        Returns:
            list: descriptors

        """
        return [d.to_json() for d in self.descriptors]

    def __reduce_ex__(self, version):
        return (
            self.__class__,
            (),
            {
                "_config": self._config,
                "_descriptors": self._descriptors,
                "_explicit_hydrogens": self._explicit_hydrogens,
                "_kekulizes": self._kekulizes,
                "_require_3D": self._require_3D,
            },
        )

    def __getitem__(self, key):
        return self._name_dict[key]

    def __init__(self, descs=None, version=None, ignore_3D=False, config=None):
        if descs is None:
            descs = []

        if config is None:
            config = {}

        self._descriptors = []
        self._name_dict = {}

        self._explicit_hydrogens = set()
        self._kekulizes = set()
        self._require_3D = False
        self._debug = False
        self._config = config

        self.register(descs, version=version, ignore_3D=ignore_3D)

    def config(self, **configs):
        r"""Set global configuration."""
        self._config.update(configs)

    @property
    def descriptors(self):
        r"""All descriptors.

        you can get/set/delete descriptor.

        Returns:
            tuple[Descriptor]: registered descriptors

        """
        return tuple(self._descriptors)

    @descriptors.setter
    def descriptors(self, descs):
        del self.descriptors
        self.register(descs)

    @descriptors.deleter
    def descriptors(self):
        self._descriptors = []
        self._name_dict = {}
        self._explicit_hydrogens.clear()
        self._kekulizes.clear()
        self._require_3D = False

    def __len__(self):
        return len(self._descriptors)

    def _register_one(self, desc, check_only=False, ignore_3D=False):
        if not isinstance(desc, Descriptor):
            raise ValueError("{!r} is not descriptor".format(desc))

        if ignore_3D and desc.require_3D:
            return

        self._explicit_hydrogens.add(bool(desc.explicit_hydrogens))
        self._kekulizes.add(bool(desc.kekulize))
        self._require_3D |= desc.require_3D

        for dep in (desc.dependencies() or {}).values():
            if isinstance(dep, Descriptor):
                self._register_one(dep, check_only=True)

        if not check_only:
            sdesc = str(desc)
            old = self._name_dict.get(sdesc)
            if old is not None:
                raise DuplicatedDescriptorName(desc, old)

            self._name_dict[sdesc] = desc
            self._descriptors.append(desc)

    def register(self, desc, version=None, ignore_3D=False):
        r"""Register descriptors.

        Descriptor-like:
            * Descriptor instance: self
            * Descriptor class: use Descriptor.preset() method
            * module: use Descriptor-likes in module
            * Iterable: use Descriptor-likes in Iterable

        Parameters:
            desc(Descriptor-like): descriptors to register
            version(str): version
            ignore_3D(bool): ignore 3D descriptors

        """
        if version is None:
            version = __version__

        version = StrictVersion(version)

        return self._register(desc, version, ignore_3D)

    def _register(self, desc, version, ignore_3D):
        if not hasattr(desc, "__iter__"):
            if is_descriptor_class(desc):
                if desc.since > version:
                    return

                for d in desc.preset(version=version):
                    self._register_one(d, ignore_3D=ignore_3D)

            elif isinstance(desc, ModuleType):
                self._register(
                    get_descriptors_in_module(desc),
                    version=version,
                    ignore_3D=ignore_3D,
                )

            else:
                self._register_one(desc, ignore_3D=ignore_3D)

        else:
            for d in desc:
                self._register(d, version=version, ignore_3D=ignore_3D)

    def _calculate_one(self, cxt, desc, reset):
        if desc in self._cache:
            return self._cache[desc]

        if reset:
            cxt.reset()
        desc._context = cxt
        cxt.add_stack(desc)

        if desc.require_connected and desc._context.n_frags != 1:
            return False, Missing(MultipleFragments(), desc._context.get_stack())

        args = {}
        for name, dep in (desc.dependencies() or {}).items():
            if dep is None:
                args[name] = None
            else:
                ok, r = self._calculate_one(cxt, dep, False)
                if ok:
                    args[name] = r
                else:
                    return False, r

        ok = False
        try:
            r = desc.calculate(**args)
            if self._debug:
                self._check_rtype(desc, r)
            ok = True
        except MissingValueException as e:
            r = Missing(e.error, desc._context.get_stack())
        except Exception as e:
            r = Error(e, desc._context.get_stack())

        self._cache[desc] = ok, r

        return ok, r

    def _check_rtype(self, desc, result):
        if desc.rtype is None:
            return

        if isinstance(result, Error):
            return

        if not isinstance(result, desc.rtype):
            raise TypeError("{} not match {}".format(result, desc.rtype))

    def _calculate(self, cxt):
        self._cache = {}
        for desc in self.descriptors:
            _, r = self._calculate_one(cxt, desc, True)
            yield r

    def __call__(self, mol, id=-1):
        r"""Calculate descriptors.

        :type mol: rdkit.Chem.Mol
        :param mol: molecular

        :type id: int
        :param id: conformer id

        :rtype: Result[scalar or Error]
        :returns: iterator of descriptor and value
        """
        return self._wrap_result(
            mol, self._calculate(Context.from_calculator(self, mol, id))
        )

    def _wrap_result(self, mol, r):
        return Result(mol, r, self._descriptors)

    def _serial(self, mols, nmols, quiet, ipynb, id):
        with self._progress(quiet, nmols, ipynb) as bar:
            for m in mols:
                with Capture() as capture:
                    r = self._wrap_result(
                        m, self._calculate(Context.from_calculator(self, m, id))
                    )

                for e in capture.result:
                    e = e.rstrip()
                    if not e:
                        continue

                    bar.write(e, file=capture.orig)

                yield r
                bar.update()

    @contextmanager
    def _progress(self, quiet, total, ipynb):
        args = {"dynamic_ncols": True, "leave": True, "total": total}

        if quiet:
            Bar = DummyBar
        elif ipynb:
            Bar = NotebookWrapper
        else:
            Bar = tqdm

        try:
            with Bar(**args) as self._progress_bar:
                yield self._progress_bar
        finally:
            if hasattr(self, "_progress_bar"):
                del self._progress_bar

    def echo(self, s, file=sys.stdout, end="\n"):
        """Output message.

        Parameters:
            s(str): message to output
            file(file-like): output to
            end(str): end mark of message

        Return:
            None

        """
        p = getattr(self, "_progress_bar", None)
        if p is not None:
            p.write(s, file=file, end="\n")
            return

        print(s, file=file, end="\n")  # noqa: T003

    def map(self, mols, nproc=None, nmols=None, quiet=False, ipynb=False, id=-1):
        r"""Calculate descriptors over mols.

        Parameters:
            mols(Iterable[rdkit.Mol]): moleculars

            nproc(int): number of process to use. default: multiprocessing.cpu_count()

            nmols(int): number of all mols to use in progress-bar. default: mols.__len__()

            quiet(bool): don't show progress bar. default: False

            ipynb(bool): use ipython notebook progress bar. default: False

            id(int): conformer id to use. default: -1.

        Returns:
            Iterator[Result[scalar]]

        """
        if nproc is None:
            nproc = cpu_count()

        if hasattr(mols, "__len__"):
            nmols = len(mols)

        if nproc == 1:
            return self._serial(mols, nmols=nmols, quiet=quiet, ipynb=ipynb, id=id)
        else:
            return self._parallel(
                mols, nproc, nmols=nmols, quiet=quiet, ipynb=ipynb, id=id
            )

    def pandas(self, mols, nproc=None, nmols=None, quiet=False, ipynb=False, id=-1):
        r"""Calculate descriptors over mols.

        Returns:
            pandas.DataFrame

        """
        from .pandas_module import MordredDataFrame, Series

        if isinstance(mols, Series):
            index = mols.index
        else:
            index = None

        return MordredDataFrame(
            (list(r) for r in self.map(mols, nproc, nmols, quiet, ipynb, id)),
            columns=[str(d) for d in self.descriptors],
            index=index,
        )


def get_descriptors_in_module(mdl, submodule=True):
    r"""Get descriptors in module.

    Parameters:
        mdl(module): module to search
        submodule(bool): search recursively

    Returns:
        Iterator[Descriptor]

    """
    __all__ = getattr(mdl, "__all__", None)
    if __all__ is None:
        __all__ = dir(mdl)

    all_values = (getattr(mdl, name) for name in __all__ if name[:1] != "_")

    if submodule:
        for v in all_values:
            if is_descriptor_class(v):
                yield v
            if isinstance(v, ModuleType):
                for v in get_descriptors_in_module(v, submodule=True):
                    yield v

    else:
        for v in all_values:
            if is_descriptor_class(v):
                yield v



from itertools import chain

import numpy as np
from six import integer_types

from ._base import Descriptor
from ._graph_matrix import DistanceMatrix, AdjacencyMatrix

__all__ = ("TopologicalCharge",)


class ChargeTermMatrix(Descriptor):
    __slots__ = ()
    explicit_hydrogens = False

    def parameters(self):
        return ()

    def dependencies(self):
        return {
            "A": AdjacencyMatrix(self.explicit_hydrogens),
            "D": DistanceMatrix(self.explicit_hydrogens),
        }

    def calculate(self, A, D):
        D2 = D.copy()
        D2[D2 != 0] **= -2
        np.fill_diagonal(D2, 0)

        M = A.dot(D2)
        return M - M.T


class TopologicalCharge(Descriptor):
    r"""topological charge descriptor.

    :type type: str
    :param type:
        * "raw": sum of order-distance atom pairs coefficient
        * "mean": mean of order-distance atom pairs coefficient
        * "global": sum of mean-topoCharge over 0 to order

    :type order: int
    :param order: int

    References
        * :doi:`10.1021/ci00019a008`

    """

    since = "1.0.0"
    __slots__ = ("_type", "_order")

    explicit_hydrogens = False

    tc_types = ("global", "mean", "raw")

    def description(self):
        return "{}-ordered {} topological charge".format(self._order, self._type)

    @classmethod
    def preset(cls, version):
        return chain(
            (cls(t, o) for t in ("raw", "mean") for o in range(1, 11)),
            [cls("global", 10)],
        )

    def __str__(self):
        if self._type == "global":
            return "JGT{}".format(self._order)
        elif self._type == "mean":
            return "JGI{}".format(self._order)
        else:
            return "GGI{}".format(self._order)

    def parameters(self):
        return self._type, self._order

    def __init__(self, type="global", order=10):
        assert type in self.tc_types
        assert type == "global" or isinstance(order, integer_types)

        self._type = type
        self._order = order

    def dependencies(self):
        return {"CT": ChargeTermMatrix(), "D": DistanceMatrix(self.explicit_hydrogens)}

    def calculate(self, CT, D):
        D = D * np.tri(*D.shape)
        D[D == 0] = np.inf

        f = D <= self._order if self._type == "global" else D == self._order

        CT = CT[f]

        if self._type == "raw":
            return np.abs(CT).sum()

        # create frequency vector
        Df = D[f]
        C = Df.copy()
        for i in np.unique(Df):
            C[Df == i] = len(Df[Df == i])

        return np.abs(CT / C).sum()

    rtype = float
